from girder import logger
from girder.api import access
from girder.api.describe import Description, autoDescribeRoute, describeRoute
from girder.api.rest import Resource, filtermodel
from girder.constants import AccessType, TokenScope
from girder.exceptions import RestException
from girder.models.item import Item
from girder.models.setting import Setting
from girder.utility.model_importer import ModelImporter

from .. import handlers
from ..constants import PluginSettings
from .system import allChildFolders, allChildItems


class HistomicsUIResource(Resource):
    def __init__(self):
        super().__init__()
        self.resourceName = 'histomicsui'

        self.route('GET', ('settings',), self.getPublicSettings)
        self.route('PUT', ('quarantine', ':id'), self.putQuarantine)
        self.route('PUT', ('quarantine', ':id', 'restore'), self.restoreQuarantine)
        # The route function tells girder to route calls to the endpoint to
        # a specific function.  The arguments here mean:
        #  1st: This is for GET requests
        #  2nd: This describes the path of the endpoint.
        #  3rd: The function that will be called.
        #
        # The `:id` component in the path is a wildcard that matches any string
        # The value matched will be passed as an argument to the function.  As
        # an example if you make a GET request to
        #   `/api/v1/histomicsui/child_metadata/foobar`
        # the function `self.getChildMetadata` will be called with the parameter
        # `id="foobar"`.
        self.route('GET', ('child_metadata', ':id'), self.getChildMetadata)

        # Similarly, this route handles calls to:
        #  `GET /api/v1/histomicsui/query_metadata`
        self.route('GET', ('query_metadata',), self.findItemsByMetadata)

    @describeRoute(
        Description('Get public settings for HistomicsUI.'),
    )
    @access.public(scope=TokenScope.DATA_READ)
    def getPublicSettings(self, params):
        keys = [
            PluginSettings.HUI_BRAND_NAME,
            PluginSettings.HUI_DEFAULT_DRAW_STYLES,
            PluginSettings.HUI_LOGIN_TEXT,
            PluginSettings.HUI_PANEL_LAYOUT,
            PluginSettings.HUI_QUARANTINE_FOLDER,
            PluginSettings.HUI_WEBROOT_PATH,
        ]
        result = {k: Setting().get(k) for k in keys}
        result[PluginSettings.HUI_QUARANTINE_FOLDER] = bool(
            result[PluginSettings.HUI_QUARANTINE_FOLDER])
        return result

    @autoDescribeRoute(
        Description('Move an item to the quarantine folder.')
        .responseClass('Item')
        .modelParam('id', model=Item, level=AccessType.WRITE)
        .errorResponse('ID was invalid.')
        .errorResponse('Write access was denied for the item', 403),
    )
    @access.user(scope=TokenScope.DATA_WRITE)
    @filtermodel(model=Item)
    def putQuarantine(self, item):
        return handlers.quarantine_item(item, self.getCurrentUser())

    @autoDescribeRoute(
        Description('Restore a quarantined item to its original folder.')
        .responseClass('Item')
        .modelParam('id', model=Item, level=AccessType.WRITE)
        .errorResponse('ID was invalid.')
        .errorResponse('Write access was denied for the item', 403),
    )
    @access.admin
    @filtermodel(model=Item)
    def restoreQuarantine(self, item):
        return handlers.restore_quarantine_item(item, self.getCurrentUser())

    # The `autoDescrbeRoute` (and `describeRoute` used in older code)
    # serves to generate the swagger documentation that looks like:
    #   https://data.kitware.com/api/v1
    #
    # The api for this is described at https://goo.gl/hnU3ws.
    @autoDescribeRoute(
        # Instantiate the instance with a basic description of the endpoint.
        Description('Get all metadata for a resource and all folders and '
                    'items that are children of a resource.')
        # Add a required "path" parameter (this is the `:id` component in the
        # route).
        .param('id', 'The ID of the resource.', paramType='path')
        # Add a required "query" parameter... e.g. `?type=collection`.
        .param('type', 'The type of the resource',
               enum=['folder', 'collection', 'user'])
        # The following to lines document common rest errors that can occur
        # when calling this endpoint.  This is for documentation only.
        .errorResponse('ID was invalid.')
        .errorResponse('Access was denied for the resource.', 403),
    )
    # This makes the endpoint accessible without logging in.
    @access.public(scope=TokenScope.DATA_READ)
    def getChildMetadata(self, id, params):
        # The `autoDescribeRoute` decorator processes the incoming request and
        # populates the function arguments.  Path parameters are added as
        # individual arguments, while query parameters are packed into the
        # `params` dictionary.
        user = self.getCurrentUser()
        modelType = params['type']
        model = ModelImporter.model(modelType)
        doc = model.load(id=id, user=user, level=AccessType.READ)
        if not doc:
            msg = 'Resource not found.'
            raise RestException(msg)
        results = {}
        if doc.get('meta'):
            results[str(doc['_id'])] = doc['meta']
        logger.info('Getting child metadata')
        for folder in allChildFolders(parentType=modelType, parent=doc,
                                      user=user, limit=0, offset=0):
            if folder.get('meta'):
                results[str(folder['_id'])] = folder['meta']
        for item in allChildItems(parentType=modelType, parent=doc,
                                  user=user, limit=0, offset=0):
            if item.get('meta'):
                results[str(item['_id'])] = item['meta']
        # By default, responses to girder endpoints are json encoded when
        # returned to the client.  In this case, it is a dictionary mapping
        # `id` -> `metadata`.
        return results

    # This endpoint returns a paginated list of all items with a given
    # (key, value) pair in their metadata.  This endpoint can be called as
    # follows:
    #
    # /histomicsui/query_metadata?
    #   key=doctor&value="John Doe"&limit=10&sort=created&sortdir=-1
    #
    # This will return a list of items matched as `{'meta': {'doctor':
    # 'John Doe'}}`.  It will return at most 10 items starting from the most
    # recently created.
    @autoDescribeRoute(
        Description('Get a list of items with a specific metadata value.')
        # This is a required string parameter representing the key in the
        # metadata.
        .param('key', 'The metadata key')
        # This is the value which should be json encoded.
        .jsonParam('value', 'The (json encoded) metadata value')
        # This adds paging parameters "limit", "offset", and "sort".  By
        # default it sorts by the `name` field of the item in ascending order.
        .pagingParams('name')
        # The following to lines document common rest errors that can occur
        # when calling this endpoint.  This is for documentation only.
        .errorResponse('Required parameters were not provided.')
        .errorResponse('Invalid value provided.'),
    )
    @access.public(scope=TokenScope.DATA_READ)
    def findItemsByMetadata(self, key, value, limit, offset, sort):
        # Construct a mongo query from the parameters given.  Developers should
        # be careful when constructing these queries to ensure that private
        # information is not leaked.  In this example, the user could pass an
        # arbitrary dictionary which could involve an aggregation pipeline, so
        # we check that only simple types are accepted.
        if isinstance(value, (list, dict)):
            # This is a special type of exception that tells girder to respond
            # with an HTTP response with the given code and message.
            msg = 'The value must not be a dictionary or list.'
            raise RestException(msg, code=400)

        query = {
            'meta': {
                key: value,
            },
        }

        # This gets the logged in user who created the request.  If it is an
        # anonymous request, this value will be `None`.
        user = self.getCurrentUser()

        # Here, item is a "model" class which is a single instance providing an
        # api that wraps traditional mongo queries.  This API is described at:
        # http://girder.readthedocs.io/en/latest/api-docs.html?#models
        item = Item()

        # This runs a "find" operation on the item collection returning a mongo
        # cursor.
        cursor = item.find(query, sort=sort)

        # The `filterResultsByPermission` allows paged access to a mongo query
        # while filtering out documents that the current user doesn't have
        # access to.  In this case, it requires the current user have read
        # access to the items.  The return value is an iterator that begins at
        # `offset` and ends at `offset + limit`.
        response = item.filterResultsByPermission(
            cursor,
            user=user, level=AccessType.READ,
            limit=limit, offset=offset,
        )

        # Finally, we turn the iterator into an explicit list for return to the
        # user.  Girder handles json encoding the response.
        return list(response)



"""Test annotation file and metadata handlers."""

import json
import time

import pytest
from girder.models.item import Item
from girder_large_image_annotation.models.annotation import Annotation

from . import girder_utilities as utilities


@pytest.mark.plugin('histomicsui')
class TestHUIHandlers:
    def testAnnotationHandler(self, server, fsAssetstore, admin):
        file = utilities.uploadExternalFile('Easy1.png', admin, fsAssetstore)
        item = Item().load(file['itemId'], user=admin)
        utilities.uploadTestFile(
            'sample.anot', admin, fsAssetstore, reference=json.dumps({
                'identifier': 'NotAnAnnotation',
                'userId': str(admin['_id']),
                'itemId': str(item['_id']),
                'fileId': str(file['_id']),
            }))
        assert Annotation().findOne({'itemId': item['_id']}) is None
        utilities.uploadTestFile(
            'sample.anot', admin, fsAssetstore, reference=json.dumps({
                'identifier': 'IsAnAnnotationFile',
                'userId': str(admin['_id']),
                'itemId': str(item['_id']),
                'fileId': str(file['_id']),
            }))
        starttime = time.time()
        while time.time() < starttime + 10:
            if Annotation().findOne({'itemId': item['_id']}) is not None:
                break
            time.sleep(0.1)
        assert Annotation().findOne({'itemId': item['_id']}) is not None

    def testMetadataHandler(self, server, fsAssetstore, admin):
        file = utilities.uploadExternalFile('Easy1.png', admin, fsAssetstore)
        item = Item().load(file['itemId'], user=admin)
        utilities.uploadTestFile(
            'sample.meta', admin, fsAssetstore, reference=json.dumps({
                'identifier': 'NotItemMetadataForItem',
                'userId': str(admin['_id']),
                'itemId': str(item['_id']),
                'fileId': str(file['_id']),
            }))
        item = Item().load(file['itemId'], user=admin)
        assert item.get('meta', {}).get('sample') is None
        utilities.uploadTestFile(
            'sample.meta', admin, fsAssetstore, reference=json.dumps({
                'identifier': 'NotItemMetadata',
                'userId': str(admin['_id']),
                'itemId': str(item['_id']),
                'fileId': str(file['_id']),
            }))
        starttime = time.time()
        while time.time() < starttime + 10:
            item = Item().load(file['itemId'], user=admin)
            if 'sample' in item.get('meta', {}):
                break
            time.sleep(0.1)
        item = Item().load(file['itemId'], user=admin)
        assert item['meta']['sample'] == 'value'
        assert item['meta']['complex']['key1'] == 'value1'

    def testAnnotationWithGirderIdHandler(self, server, fsAssetstore, admin):
        file = utilities.uploadExternalFile('Easy1.png', admin, fsAssetstore)
        item = Item().load(file['itemId'], user=admin)
        utilities.uploadExternalFile(
            'Easy1.png', admin, fsAssetstore, reference=json.dumps({
                'identifier': 'ImageRecord1',
                'uuid': '12345',
                'userId': str(admin['_id']),
                'itemId': str(item['_id']),
                'fileId': str(file['_id']),
            }))
        assert Annotation().findOne({'itemId': item['_id']}) is None
        utilities.uploadTestFile(
            'sample_girder_id.anot', admin, fsAssetstore, reference=json.dumps({
                'identifier': 'IsAnAnnotationFile',
                'uuid': '12345',
                'userId': str(admin['_id']),
                'itemId': str(item['_id']),
                'fileId': str(file['_id']),
            }))
        starttime = time.time()
        while time.time() < starttime + 10:
            if Annotation().findOne({'itemId': item['_id']}) is not None:
                break
            time.sleep(0.1)
        assert Annotation().findOne({'itemId': item['_id']}) is not None

    def testAnnotationWithGirderIdHandlerAltOrder(self, server, fsAssetstore, admin):
        file = utilities.uploadExternalFile('Easy1.png', admin, fsAssetstore)
        item = Item().load(file['itemId'], user=admin)
        utilities.uploadTestFile(
            'sample_girder_id.anot', admin, fsAssetstore, reference=json.dumps({
                'identifier': 'IsAnAnnotationFile',
                'uuid': '12346',
                'userId': str(admin['_id']),
                'itemId': str(item['_id']),
                'fileId': str(file['_id']),
            }))
        assert Annotation().findOne({'itemId': item['_id']}) is None
        utilities.uploadExternalFile(
            'Easy1.png', admin, fsAssetstore, reference=json.dumps({
                'identifier': 'ImageRecord1',
                'uuid': '12346',
                'userId': str(admin['_id']),
                'itemId': str(item['_id']),
                'fileId': str(file['_id']),
            }))
        starttime = time.time()
        while time.time() < starttime + 10:
            if Annotation().findOne({'itemId': item['_id']}) is not None:
                break
            time.sleep(0.1)
        assert Annotation().findOne({'itemId': item['_id']}) is not None



from girder.models.file import File
from girder.models.item import Item
from girder.models.model_base import ValidationException
from girder_large_image.models.image_item import ImageItem

from .meta import TCGAModel
from .slide import Slide


class Image(TCGAModel, Item):

    TCGAType = 'image'

    def validate(self, doc, **kwargs):
        super().validate(doc, **kwargs)
        if 'largeImage' not in doc:
            msg = 'An image item must be a "large_image"'
            raise ValidationException(
                msg,
            )
        slide = Slide().load(
            doc['folderId'], force=True)
        if not slide or self.getTCGAType(slide) != 'slide':
            msg = 'An image must be a child of a slide'
            raise ValidationException(
                msg,
            )
        tcga = self.getTCGA(doc)
        if not self.case_re.match(tcga.get('case', '')):
            msg = 'Invalid case name in TCGA metadata'
            raise ValidationException(
                msg,
            )
        if not self.barcode_re.match(tcga.get('barcode', '')):
            msg = 'Invalid barcode in TCGA metadata'
            raise ValidationException(
                msg,
            )
        if not self.uuid_re.match(tcga.get('uuid', '')):
            msg = 'Invalid uuid in TCGA metadata'
            raise ValidationException(
                msg,
            )

        return doc

    def _setLargeImage(self, doc, fileId, user, token):
        if doc.get('largeImage', {}).get('fileId') == fileId:
            return
        file = File().load(fileId, user=user)
        try:
            return ImageItem.createImageItem(
                doc, file,
                user=user, token=token, createJob=False,
            )
        except Exception:
            msg = 'Could not generate a large_image item'
            raise ValidationException(
                msg,
            )

    def _findImageFile(self, doc):
        for file in self.childFiles(doc):
            if self.image_re.match(file['name']):
                return file['_id']

    def importDocument(self, doc, **kwargs):
        """Import a slide item into a `case` folder."""
        user = kwargs.get('user')
        token = kwargs.get('token')
        fileId = self._findImageFile(doc)
        if fileId is None:
            msg = 'Could not find a TCGA slide in item'
            raise ValidationException(
                msg,
            )
        self._setLargeImage(doc, fileId, user, token)

        name = doc['name']
        parent = Slide().load(
            doc.get('folderId'), force=True,
        )
        if not parent:
            msg = 'Invalid item document'
            raise ValidationException(
                msg,
            )
        tcga = self.getTCGA(parent)
        tcga.update(self.parseImage(name))
        self.setTCGA(doc, **tcga)
        return super().importDocument(doc, **kwargs)



import girder_large_image
from girder.api import access
from girder.api.describe import Description, autoDescribeRoute
from girder.api.v1.item import Item as ItemResource
from girder.constants import AccessType, TokenScope
from girder.exceptions import RestException
from girder.models.folder import Folder


def _isLargeImageItem(item):
    return item.get('largeImage', {}).get('fileId') is not None


class ImageBrowseResource(ItemResource):
    """Extends the "item" resource to iterate through images im a folder."""

    def __init__(self, apiRoot):
        # Don't call the parent (Item) constructor, to avoid redefining routes,
        # but do call the grandparent (Resource) constructor
        super(ItemResource, self).__init__()

        self.resourceName = 'item'
        apiRoot.item.route('GET', (':id', 'next_image'), self.getNextImage)
        apiRoot.item.route('GET', (':id', 'previous_image'), self.getPreviousImage)
        apiRoot.item.route('GET', (':id', 'adjacent_images'), self.getPreviousAndNextImages)

    def getAdjacentImages(self, currentImage, currentFolder=None):
        folderModel = Folder()
        if currentFolder:
            folder = currentFolder
        else:
            folder = folderModel.load(
                currentImage['folderId'], user=self.getCurrentUser(), level=AccessType.READ)

        sort = [('name', 1)]
        try:
            conf = girder_large_image.YAMLConfigFile(
                folder, '.large_image_config.yaml', self.getCurrentUser())
            if conf['itemList']['defaultSort']:
                sort = [(
                    ('meta.' if entry['type'] == 'metadata' else '') + entry['value'],
                    1 if entry['dir'] == 'down' else -1)
                    for entry in conf['itemList']['defaultSort']]
        except Exception:
            pass

        if folder.get('isVirtual'):
            children = folderModel.childItems(folder, sort=sort, includeVirtual=True)
        else:
            children = folderModel.childItems(folder, sort=sort)

        allImages = [item for item in children if _isLargeImageItem(item)]
        try:
            index = allImages.index(currentImage)
        except ValueError:
            msg = 'Id is not an image'
            raise RestException(msg, 404)

        return {
            'previous': allImages[index - 1],
            'next': allImages[(index + 1) % len(allImages)],
            'index': index,
            'count': len(allImages),
        }

    @access.public(scope=TokenScope.DATA_READ)
    @autoDescribeRoute(
        Description('Get the next image in the same folder as the given item.')
        .modelParam('id', 'The current image ID',
                    model='item', destName='image', paramType='path', level=AccessType.READ)
        .modelParam('folderId', 'The (virtual) folder ID the image is located in',
                    model='folder', destName='folder', paramType='query', level=AccessType.READ,
                    required=False)
        .errorResponse()
        .errorResponse('Image not found', code=404),
    )
    def getNextImage(self, image, folder):
        return self.getAdjacentImages(image, folder)['next']

    @access.public(scope=TokenScope.DATA_READ)
    @autoDescribeRoute(
        Description('Get the previous image in the same folder as the given item.')
        .modelParam('id', 'The current item ID',
                    model='item', destName='image', paramType='path', level=AccessType.READ)
        .modelParam('folderId', 'The (virtual) folder ID the image is located in',
                    model='folder', destName='folder', paramType='query', level=AccessType.READ,
                    required=False)
        .errorResponse()
        .errorResponse('Image not found', code=404),
    )
    def getPreviousImage(self, image, folder):
        return self.getAdjacentImages(image, folder)['previous']

    @access.public(scope=TokenScope.DATA_READ)
    @autoDescribeRoute(
        Description('Get the previous and next image in the same folder as the given item.')
        .modelParam('id', 'The current item ID',
                    model='item', destName='image', paramType='path', level=AccessType.READ)
        .modelParam('folderId', 'The (virtual) folder ID the image is located in',
                    model='folder', destName='folder', paramType='query', level=AccessType.READ,
                    required=False)
        .errorResponse()
        .errorResponse('Image not found', code=404),
    )
    def getPreviousAndNextImages(self, image, folder):
        return self.getAdjacentImages(image, folder)



from girder.models.folder import Folder
from girder.models.model_base import ValidationException

from .case import Case
from .meta import TCGAModel


class Slide(TCGAModel, Folder):

    TCGAType = 'slide'

    def validate(self, doc, **kwargs):
        if not doc.get('parentCollection') == 'folder':
            msg = 'A Slide model must be a child of a folder'
            raise ValidationException(
                msg,
            )
        super().validate(doc, **kwargs)
        case = Case().load(
            doc['parentId'], force=True)
        if not case or self.getTCGAType(case) != 'case':
            msg = 'A Slide model must be a child of a case'
            raise ValidationException(
                msg,
            )
        return doc

    def importDocument(self, doc, **kwargs):
        from .image import Image

        recurse = kwargs.get('recurse', False)
        parent = Case().load(
            doc.get('parentId'), force=True,
        )
        if not parent:
            msg = 'Invalid folder document'
            raise ValidationException(
                msg,
            )
        tcga = self.getTCGA(parent)
        tcga.pop('meta', None)
        self.setTCGA(doc, **tcga)
        doc = super().importDocument(doc, **kwargs)
        if not recurse:
            return doc

        childModel = Image()
        children = Folder().childItems(
            doc, user=kwargs.get('user'),
        )
        for child in children:
            try:
                childModel.importDocument(child, **kwargs)
            except ValidationException:
                pass
        return doc



from . import aperio, system, tcga
from .hui_resource import HistomicsUIResource
from .image_browse_resource import ImageBrowseResource


def addEndpoints(apiRoot):
    """
    This adds endpoints from each module.

    :param apiRoot: Girder api root class.
    """
    system.addSystemEndpoints(apiRoot)
    apiRoot.tcga = tcga.TCGAResource()
    aperio.addItemEndpoints(apiRoot.item)
    aperio.addTcgaEndpoints(apiRoot.tcga)

    ImageBrowseResource(apiRoot)

    apiRoot.histomicsui = HistomicsUIResource()



#############################################################################
#  Copyright Kitware Inc.
#
#  Licensed under the Apache License, Version 2.0 ( the "License" );
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#############################################################################

import json
import logging
import os
import re
from functools import wraps

import cherrypy
from bson import json_util
from girder import events, logger, plugin
from girder.api import access
from girder.api.rest import getCurrentToken
from girder.constants import AssetstoreType
from girder.exceptions import AccessException, ValidationException
from girder.models.assetstore import Assetstore
from girder.models.file import File
from girder.models.folder import Folder
from girder.models.item import Item
from girder.models.setting import Setting
from girder.models.token import Token
from girder.settings import SettingDefault, SettingKey
from girder.utility import config
from girder.utility import path as path_util
from girder.utility import setting_utilities
from girder.utility.model_importer import ModelImporter
from girder.utility.webroot import Webroot
from pkg_resources import DistributionNotFound, get_distribution

from . import handlers, rest
from .constants import PluginSettings
from .models.aperio import Aperio
from .models.case import Case
from .models.cohort import Cohort
from .models.image import Image
from .models.pathology import Pathology
from .models.slide import Slide

try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    # package is not installed
    pass

# There are other packages that add to the root log handler; this makes the
# system very noisy.  Stop that.
while logging.root.hasHandlers():
    logging.root.removeHandler(logging.root.handlers[0])
logging.root.addHandler(logging.NullHandler())

_template = os.path.join(
    os.path.dirname(__file__),
    'webroot.mako',
)


def patchCookieParsing():
    """
    Python's http cookie parser fails for all cookies when there are some
    invalid cookies.  Work around some of that.

    See https://bugs.python.org/issue31456 for a discussion.
    See https://github.com/python/cpython/blob/master/Lib/http/cookies.py for
    the relevant Python source code.
    """
    try:
        # This will fail in Python 2.7.  Python 2.7 has a backport module, but
        # it is constructed differently.
        import http.cookies

        # This is both a sanity check and to make sure we don't do things
        # multiple times.  It should work on python 3.5 - 3.9 (and possibly
        # later).  See
        if (r'\s' not in http.cookies._LegalValueChars and
                r'\w' in http.cookies._LegalValueChars and
                http.cookies.BaseCookie._BaseCookie__parse_string.__defaults__ == (
                    http.cookies._CookiePattern, )):
            http.cookies._LegalValueChars += r'\s'
            http.cookies._CookiePattern = re.compile(
                http.cookies._CookiePattern.pattern.replace(r'\[\]', r'\[\]\s'),
                http.cookies._CookiePattern.flags)
            http.cookies.BaseCookie._BaseCookie__parse_string.__defaults__ = (
                http.cookies._CookiePattern, )
    except Exception:
        pass


def betterInvalidateJSandCSSCaches(root):
    from girder import constants

    origRenderHTML = root._renderHTML

    def _renderHTML(self):
        topBuiltDir = os.path.join(constants.STATIC_ROOT_DIR, 'built')
        result = origRenderHTML()
        lastUpdate = 0
        for filename in {
                'girder_lib.min.js', 'girder_app.min.js', 'girder_lib.min.css',
                'Girder_Favicon.png'}:
            filepath = os.path.join(topBuiltDir, filename)
            if os.path.exists(filepath):
                lastUpdate = max(lastUpdate, os.path.getmtime(filepath))
        builtDir = os.path.join(constants.STATIC_ROOT_DIR, 'built', 'plugins')
        for pluginName in self.vars['plugins']:
            for filepath in [
                os.path.join(builtDir, pluginName, 'plugin.min.css'),
                os.path.join(builtDir, pluginName, 'plugin.min.js'),
            ]:
                if os.path.exists(filepath):
                    lastUpdate = max(lastUpdate, os.path.getmtime(filepath))
        luParam = '?_=%d' % int(lastUpdate * 1000)
        while True:
            match = re.search(
                r'<(link rel="(icon|stylesheet)" [^>]*href="[^>?"]*\.(css|png)|'
                r'script src="[^>?"]*.js)">', result)
            if not match:
                break
            result = result[:match.span()[1] - 2] + luParam + result[match.span()[1] - 2:]
        return result

    root._renderHTML = _renderHTML.__get__(root)


@setting_utilities.validator({
    PluginSettings.HUI_DEFAULT_DRAW_STYLES,
    PluginSettings.HUI_PANEL_LAYOUT,
})
def validateListOrJSON(doc):
    val = doc['value']
    try:
        if isinstance(val, list):
            doc['value'] = json.dumps(val)
        elif val is None or val.strip() == '':
            doc['value'] = None
        else:
            parsed = json.loads(val)
            if not isinstance(parsed, list):
                raise ValueError
            doc['value'] = val.strip()
    except (ValueError, AttributeError):
        raise ValidationException('%s must be a JSON list.' % doc['key'], 'value')


@setting_utilities.validator({
    PluginSettings.HUI_BANNER_COLOR,
    PluginSettings.HUI_BRAND_COLOR,
})
def validateHistomicsUIColor(doc):
    if not doc['value']:
        msg = 'The banner color may not be empty'
        raise ValidationException(msg, 'value')
    elif not re.match(r'^#[0-9A-Fa-f]{6}$', doc['value']):
        msg = 'The banner color must be a hex color triplet'
        raise ValidationException(msg, 'value')


@setting_utilities.validator(PluginSettings.HUI_BRAND_NAME)
def validateHistomicsUIBrandName(doc):
    if not doc['value']:
        msg = 'The brand name may not be empty'
        raise ValidationException(msg, 'value')


@setting_utilities.validator(PluginSettings.HUI_WEBROOT_PATH)
def validateHistomicsUIWebrootPath(doc):
    if not doc['value']:
        msg = 'The webroot path may not be empty'
        raise ValidationException(msg, 'value')
    if re.match(r'^girder$', doc['value']):
        msg = 'The webroot path may not be "girder"'
        raise ValidationException(msg, 'value')


@setting_utilities.validator(PluginSettings.HUI_ALTERNATE_WEBROOT_PATH)
def validateHistomicsUIAlternateWebrootPath(doc):
    if re.match(r'(^,|)girder(,|$)', doc['value']):
        msg = 'The alternate webroot path may not contain "girder"'
        raise ValidationException(msg, 'value')


@setting_utilities.validator(PluginSettings.HUI_QUARANTINE_FOLDER)
def validateHistomicsUIQuarantineFolder(doc):
    if not doc.get('value', None):
        doc['value'] = None
    else:
        Folder().load(doc['value'], force=True, exc=True)


@setting_utilities.validator(PluginSettings.HUI_LOGIN_SESSION_EXPIRY_MINUTES)
def validateLoginSessionExpiryMinutes(doc):
    if not doc.get('value', None):
        doc['value'] = None
    else:
        try:
            doc['value'] = float(doc['value'])
            if doc['value'] > 0:
                return
        except ValueError:
            pass  # We want to raise the ValidationException
        msg = 'Login session expiry must be None or a number >= 0.0.'
        raise ValidationException(msg, 'value')


# Defaults that have fixed values are added to the system defaults dictionary.
SettingDefault.defaults.update({
    PluginSettings.HUI_WEBROOT_PATH: 'histomics',
    PluginSettings.HUI_BRAND_NAME: 'HistomicsUI',
    PluginSettings.HUI_BANNER_COLOR: '#f8f8f8',
    PluginSettings.HUI_BRAND_COLOR: '#777777',
    PluginSettings.HUI_HELP_URL:
        'https://github.com/DigitalSlideArchive/HistomicsUI/blob/master/docs/controls.rst',
    PluginSettings.HUI_HELP_TOOLTIP: 'Mouse and keyboard controls',
    PluginSettings.HUI_HELP_TEXT: 'Help',
})


@setting_utilities.validator({
    PluginSettings.HUI_HELP_URL,
    PluginSettings.HUI_HELP_TOOLTIP,
    PluginSettings.HUI_HELP_TEXT,
    PluginSettings.HUI_LOGIN_TEXT,
})
def validateHistomicsUIHelp(doc):
    pass


@setting_utilities.validator({
    PluginSettings.HUI_DELETE_ANNOTATIONS_AFTER_INGEST,
})
def validateBoolean(doc):
    val = doc['value']
    if str(val).lower() not in ('false', 'true', ''):
        raise ValidationException('%s must be a boolean.' % doc['key'], 'value')
    doc['value'] = (str(val).lower() == 'true')


def childItems(self, folder, limit=0, offset=0, sort=None, filters=None,
               includeVirtual=False, **kwargs):
    if not includeVirtual or not folder.get('isVirtual') or 'virtualItemsQuery' not in folder:
        return Folder._childItemsBeforeHUI(
            self, folder, limit=limit, offset=offset, sort=sort,
            filters=filters, **kwargs)
    q = json_util.loads(folder['virtualItemsQuery'])
    if 'virtualItemsSort' in folder and sort is None:
        sort = json.loads(folder['virtualItemsSort'])
    q.update(filters or {})
    return Item().find(q, limit=limit, offset=offset, sort=sort, **kwargs)


def _saveJob(event):
    """
    When a job is saved, if it is a docker run task, add the Dask Bokeh port to
    the list of exposed ports.
    """
    job = event.info
    try:
        jobkwargs = json_util.loads(job['kwargs'])
        if '--scheduler' in jobkwargs['container_args']:
            jobkwargs['ports'] = {'8787': None}
            job['kwargs'] = json_util.dumps(jobkwargs)
    except Exception:
        pass


class WebrootHistomicsUI(Webroot):
    def _renderHTML(self):
        self.updateHtmlVars({
            'title': Setting().get(PluginSettings.HUI_BRAND_NAME),
            'huiBrandName': Setting().get(PluginSettings.HUI_BRAND_NAME),
            'huiBrandColor': Setting().get(PluginSettings.HUI_BRAND_COLOR),
            'huiBannerColor': Setting().get(PluginSettings.HUI_BANNER_COLOR),
            'huiHelpURL': Setting().get(PluginSettings.HUI_HELP_URL),
            'huiHelpTooltip': Setting().get(PluginSettings.HUI_HELP_TOOLTIP),
            'huiHelpText': Setting().get(PluginSettings.HUI_HELP_TEXT),
        })
        return super()._renderHTML()

    def GET(self, **params):
        print(params)
        if params.get('token'):
            try:
                token = Token().load(params['token'], force=True, objectId=False)
                if token:
                    cookie = cherrypy.response.cookie
                    cookie['girderToken'] = str(token['_id'])
                    cookie['girderToken']['path'] = '/'
                    days = float(Setting().get(SettingKey.COOKIE_LIFETIME))
                    cookie['girderToken']['expires'] = int(days * 3600 * 24)
                    cookie['girderToken']['sameSite'] = 'None'
                    cookie['girderToken']['secure'] = True
            except Exception:
                pass
        return self._renderHTML()


def restrict_downloads(info):
    """
    If restrict_downloads is configured, modify endpoints to do that.
    """
    curConfig = config.getConfig().get('histomicsui', {})
    if not curConfig.get('restrict_downloads'):
        return
    # Change some endpoints to require token access
    endpoints = [
        ('collection', 'GET', (':id', 'download'), True),
        ('file', 'GET', (':id', 'download', ':name'), True),
        ('folder', 'GET', (':id', 'download'), True),
        ('resource', 'GET', ('download', ), False),
        ('resource', 'POST', ('download', ), True),

        ('item', 'GET', (':itemId', 'tiles', 'images', ':image'), True),
    ]
    intEndpoints = [
        ('file', 'GET', (':id', 'download'), True),
        ('item', 'GET', (':id', 'download'), True),
    ]
    if not isinstance(curConfig['restrict_downloads'], int):
        endpoints += intEndpoints

    for resource, method, route, clrScope in endpoints:
        cls = getattr(info['apiRoot'], resource)
        boundfunc = cls.getRouteHandler(method, route)
        func = getattr(boundfunc, '__func__', boundfunc)
        if func.accessLevel == 'public':
            newfunc = access.token(func)
            newfunc.requiredScopes = getattr(func, 'requiredScopes', None)
            if getattr(func, 'requiredScopes', None) and clrScope:
                del func.requiredScopes
            if getattr(func, 'cookieAuth', None) and clrScope:
                newfunc.cookieAuth = True
                del func.cookieAuth
            # Rebind new function
            if boundfunc != func:
                newfunc = newfunc.__get__(boundfunc.__self__, boundfunc.__class__)
                setattr(newfunc.__self__, newfunc.__name__, newfunc)
            cls.removeRoute(method, route)
            cls.route(method, route, newfunc)
    if isinstance(curConfig['restrict_downloads'], int):
        def limitLengthDownload(fun, limit):
            @wraps(fun)
            def wrapped(*args, **kwargs):
                if (not getCurrentToken() and len(args) >= 1 and
                        args[0] and args[0].get('size', 0) >= limit):
                    msg = 'You must be logged in or have a valid auth token.'
                    raise AccessException(
                        msg)
                return fun(*args, **kwargs)
            return wrapped
        File().download = limitLengthDownload(
            File().download, curConfig['restrict_downloads'])


def cleanupFSAssetstores():
    curConfig = config.getConfig().get('histomicsui', {})
    if not curConfig.get('cleanup_fsassetstore'):
        return
    check = curConfig.get('cleanup_fsassetstore') == 'check'
    for assetstore in Assetstore().find({'type': AssetstoreType.FILESYSTEM}):
        fullcount = 0
        count = 0
        total = 0
        base = assetstore['root'].rstrip(os.sep) + os.sep
        logger.info('Checking assetstore if it needs cleanup: %s at %s', assetstore['name'], base)
        paths = {entry['path'] for entry in File().find({
            'assetstoreId': assetstore['_id'], 'imported': {'$exists': False}})}
        for root, _dirs, files in os.walk(base):
            for file in files:
                fullcount += 1
                path = os.path.join(root, file)
                if path.startswith(base) and path[len(base):] not in paths:
                    count += 1
                    total += os.path.getsize(path)
                    if not check:
                        logger.info('Unlink unused file: %s (%d)',
                                    path, os.path.getsize(path))
                        os.unlink(path)
                    else:
                        logger.info('Would unlink unused file: %s (%d)',
                                    path, os.path.getsize(path))
        logger.info('Finished checking assetstore: %s (%d/%d file(s), %d byte(s))',
                    assetstore['name'], count, fullcount, total)


class GirderPlugin(plugin.GirderPlugin):
    DISPLAY_NAME = 'HistomicsUI'
    CLIENT_SOURCE_PATH = 'web_client'

    def load(self, info):  # noqa
        plugin.getPlugin('jobs').load(info)
        try:
            plugin.getPlugin('slicer_cli_web').load(info)
        except Exception:
            logger.info('slicer_cli_web plugin is unavailable.  Analysis '
                        'tasks are therefore unavailable.')
        plugin.getPlugin('large_image_annotation').load(info)

        # Support short login sessions
        handlers.shortLoginSessions()
        # Python's http cookie parser fails for all cookies when there are some
        # invalid cookies.  Work around some of that.
        patchCookieParsing()

        ModelImporter.registerModel('aperio', Aperio, 'histomicsui')
        ModelImporter.registerModel('case', Case, 'histomicsui')
        ModelImporter.registerModel('cohort', Cohort, 'histomicsui')
        ModelImporter.registerModel('image', Image, 'histomicsui')
        ModelImporter.registerModel('pathology', Pathology, 'histomicsui')
        ModelImporter.registerModel('slide', Slide, 'histomicsui')

        rest.addEndpoints(info['apiRoot'])
        info['serverRoot'].updateHtmlVars({
            'brandName': Setting().get(SettingKey.BRAND_NAME)})
        # Better virtual folder support
        if not getattr(Folder, '_childItemsBeforeHUI', None):
            Folder._childItemsBeforeHUI = Folder.childItems
            Folder.childItems = childItems
        # Allow virtual folders to form resource paths.  This could be moved to
        # the virtual folder plugin's load method.
        oldLookUpToken = path_util.lookUpToken

        def lookUpToken(token, parentType, parent):
            if parentType == 'folder' and parent.get('isVirtual') and 'virtualItemsQuery' in parent:
                q = json_util.loads(parent['virtualItemsQuery'])
                q = {'$and': [q, {'name': token}]}
                item = Item().findOne(q)
                if item:
                    return item, 'item'
            return oldLookUpToken(token, parentType, parent)

        path_util.lookUpToken = lookUpToken

        girderRoot = info['serverRoot']
        huiRoot = WebrootHistomicsUI(_template)
        huiRoot.updateHtmlVars(girderRoot.vars)

        betterInvalidateJSandCSSCaches(girderRoot)
        betterInvalidateJSandCSSCaches(huiRoot)

        # The interface is always available under hui and also available
        # under the specified path.
        info['serverRoot'].hui = huiRoot
        webrootPath = Setting().get(PluginSettings.HUI_WEBROOT_PATH)
        alternateWebrootPath = Setting().get(PluginSettings.HUI_ALTERNATE_WEBROOT_PATH)
        setattr(info['serverRoot'], webrootPath, huiRoot)
        if alternateWebrootPath:
            for alt_webroot_path in alternateWebrootPath.split(','):
                if alt_webroot_path:
                    setattr(info['serverRoot'], alt_webroot_path, huiRoot)
        info['serverRoot'].girder = girderRoot

        # Auto-ingest annotations into database when a file with an identifier
        # ending in 'AnnotationFile' is uploaded (usually .anot files).
        events.bind('data.process', 'histomicsui.annotations', handlers.process_annotations)
        # Auto-ingest metadata into parent when a file with an identifier
        # ending in 'ItemMetadata' is uploaded (usually .meta files).
        events.bind('data.process', 'histomicsui.metadata', handlers.process_metadata)

        events.bind('model.job.save', 'histomicsui', _saveJob)

        handlers.json_nans_as_nulls()

        def updateWebroot(event):
            """
            If the webroot path setting is changed, bind the new path to the
            hui webroot resource.  Note that a change to the alternate webroot
            requires a restart.
            """
            if event.info.get('key') == PluginSettings.HUI_WEBROOT_PATH:
                setattr(info['serverRoot'], event.info['value'], huiRoot)

        events.bind('model.setting.save.after', 'histomicsui', updateWebroot)

        restrict_downloads(info)
        cleanupFSAssetstores()



from girder.settings import SettingDefault, SettingKey

TCGACollectionSettingKey = 'tcga.tcga_collection_id'

SettingDefault.defaults[SettingKey.BRAND_NAME] = 'HistomicsUI'


# Constants representing the setting keys for this plugin
class PluginSettings:
    HUI_DEFAULT_DRAW_STYLES = 'histomicsui.default_draw_styles'
    HUI_PANEL_LAYOUT = 'histomicsui.panel_layout'
    HUI_WEBROOT_PATH = 'histomicsui.webroot_path'
    HUI_ALTERNATE_WEBROOT_PATH = 'histomicsui.alternate_webroot_path'
    HUI_BRAND_NAME = 'histomicsui.brand_name'
    HUI_BRAND_COLOR = 'histomicsui.brand_color'
    HUI_BANNER_COLOR = 'histomicsui.banner_color'
    HUI_QUARANTINE_FOLDER = 'histomicsui.quarantine_folder'
    HUI_DELETE_ANNOTATIONS_AFTER_INGEST = 'histomicsui.delete_annotations_after_ingest'
    HUI_HELP_URL = 'histomicsui.help_url'
    HUI_HELP_TOOLTIP = 'histomicsui.help_tooltip'
    HUI_HELP_TEXT = 'histomicsui.help_text'
    HUI_LOGIN_TEXT = 'histomicsui.login_text'
    HUI_LOGIN_SESSION_EXPIRY_MINUTES = 'histomicsui.login_session_expiry_minutes'



from girder.models.folder import Folder
from girder.models.model_base import ValidationException

from .cohort import Cohort
from .meta import TCGAModel


class Case(TCGAModel, Folder):

    TCGAType = 'case'
    TCGAIndices = [
        'tcga.label',
    ]

    def validate(self, doc, **kwargs):
        if doc.get('parentCollection') != 'folder':
            msg = 'A Case model must be a child of a folder'
            raise ValidationException(
                msg,
            )
        super().validate(doc, **kwargs)
        cohort = Cohort().load(
            doc['parentId'], force=True)
        if not cohort or self.getTCGAType(cohort) != 'cohort':
            msg = 'A Case model must be a child of a cohort'
            raise ValidationException(
                msg,
            )
        if not self.case_re.match(self.getTCGA(doc).get('label', '')):
            msg = 'Invalid label in TCGA metadata'
            raise ValidationException(
                msg,
            )
        return doc

    def importDocument(self, doc, **kwargs):
        from .slide import Slide

        recurse = kwargs.get('recurse', False)
        parent = Cohort().load(
            doc.get('parentId'), force=True,
        )
        if not parent:
            msg = 'Invalid folder document'
            raise ValidationException(
                msg,
            )
        tcga = self.getTCGA(parent)
        tcga['label'] = doc['name']
        tcga['caseId'] = doc['_id']
        self.setTCGA(doc, **tcga)
        doc = super().importDocument(doc, **kwargs)
        if not recurse:
            return doc

        childModel = Slide()
        children = Folder().childFolders(
            doc, 'folder', user=kwargs.get('user'),
        )
        for child in children:
            try:
                childModel.importDocument(child, **kwargs)
            except ValidationException:
                pass
        return doc



from girder.models.item import Item
from girder.models.model_base import ValidationException

from .meta import TCGAModel


class Pathology(TCGAModel, Item):

    TCGAType = 'pathology'

    def importDocument(self, doc, **kwargs):
        recurse = kwargs.get('recurse', False)

        if doc.get('_modelType', 'item') == 'item':
            name = doc['name']
            tcga = self.parsePathology(name)
            self.setTCGA(doc, **tcga)

            doc = super().importDocument(doc, **kwargs)
            return doc
        elif recurse:
            for item in self.iterateItems(doc):
                try:
                    self.importDocument(item, **kwargs)
                except ValidationException:
                    pass
        else:
            msg = 'Invalid model type'
            raise ValidationException(msg)



#############################################################################
#  Copyright Kitware Inc.
#
#  Licensed under the Apache License, Version 2.0 ( the "License" );
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#############################################################################

import datetime
import os

from girder import logger
from girder.api import access
from girder.api.describe import Description, autoDescribeRoute, describeRoute
from girder.api.rest import RestException, boundHandler, filtermodel
from girder.api.v1.resource import Resource as ResourceResource
from girder.constants import AccessType, AssetstoreType, TokenScope
from girder.models.assetstore import Assetstore
from girder.models.file import File
from girder.models.folder import Folder
from girder.models.item import Item
from girder.models.setting import Setting
from girder.utility import path as path_util
from girder.utility.model_importer import ModelImporter
from girder_jobs.models.job import Job


def addSystemEndpoints(apiRoot):
    """
    This adds endpoints to routes that already exist in Girder.

    :param apiRoot: Girder api root class.
    """
    # Added to the item route
    apiRoot.item.route('GET', ('query',), getItemsByQuery)
    # Added to the folder route
    apiRoot.folder.route('GET', ('query',), getFoldersByQuery)
    # Added to the file route
    apiRoot.file.route('GET', ('query',), getFilesByQuery)
    # Added to the system route
    apiRoot.system.route('PUT', ('restart',), restartServer)
    apiRoot.system.route('GET', ('setting', 'default'), getSettingDefault)
    # Added to the job route
    apiRoot.job.route('GET', ('old',), getOldJobs)
    apiRoot.job.route('DELETE', ('old',), deleteOldJobs)
    # Added to the file route
    apiRoot.file.route('POST', (':id', 'import', 'adjust_path'), adjustFileImportPath)
    # Added to the resource route
    apiRoot.resource.route('POST', ('paths',), getMultipleResourcePaths)
    # Added to the histomicui route
    HUIResourceResource(apiRoot)

    # Allow getting plugin resources
    origGetResourceModel = ResourceResource._getResourceModel

    def _getResourceModel(self, kind, funcName=None):
        model = None
        try:
            model = origGetResourceModel(self, kind, funcName=funcName)
        except Exception:
            if model is None and '.' in str(kind):
                try:
                    model = ModelImporter.model(
                        kind.split('.', 1)[-1], plugin=kind.split('.', 1)[0])
                except Exception:
                    model = None
            if not model or (funcName and not hasattr(model, funcName)):
                msg = 'Invalid resources format.'
                raise RestException(msg)
        return model

    ResourceResource._getResourceModel = _getResourceModel

    # Allow users to see what user created a job
    Job().exposeFields(level=AccessType.READ, fields={'userId'})


def allChildFolders(parent, parentType, user, limit=0, offset=0,
                    sort=None, _internal=None, **kwargs):
    """
    This generator will yield all folders that are children of the resource
    or recursively children of child folders of the resource, with access
    policy filtering.  Passes any kwargs to the find function.

    :param parent: The parent object.
    :type parentType: Type of the parent object.
    :param parentType: The parent type.
    :type parentType: 'user', 'folder', or 'collection'
    :param user: The user running the query.  Only returns folders that this
                 user can see.
    :param limit: Result limit.
    :param offset: Result offset.
    :param sort: The sort structure to pass to pymongo.  Child folders are
        served depth first, and this sort is applied within the resource
        and then within each child folder.
    """
    if _internal is None:
        _internal = {
            'limit': limit,
            'offset': offset,
            'done': False,
        }
    for folder in Folder().childFolders(
            parentType=parentType, parent=parent, user=user,
            limit=_internal['limit'], offset=0, sort=sort, **kwargs):
        if _internal['done']:
            return
        if _internal['offset']:
            _internal['offset'] -= 1
        else:
            yield folder
            if _internal['limit']:
                _internal['limit'] -= 1
                if not _internal['limit']:
                    _internal['done'] = True
                    return
        yield from allChildFolders(
            folder, 'folder', user, sort=sort, _internal=_internal, **kwargs)


def allChildItems(parent, parentType, user, limit=0, offset=0,
                  sort=None, _internal=None, **kwargs):
    """
    This generator will yield all items that are children of the resource
    or recursively children of child folders of the resource, with access
    policy filtering.  Passes any kwargs to the find function.

    :param parent: The parent object.
    :type parentType: Type of the parent object.
    :param parentType: The parent type.
    :type parentType: 'user', 'folder', or 'collection'
    :param user: The user running the query. Only returns items that this
                 user can see.
    :param limit: Result limit.
    :param offset: Result offset.
    :param sort: The sort structure to pass to pymongo.  Child folders are
        served depth first, and this sort is applied within the resource
        and then within each child folder.  Child items are processed
        before child folders.
    """
    if _internal is None:
        _internal = {
            'limit': limit,
            'offset': offset,
            'done': False,
        }
    model = ModelImporter.model(parentType)
    if hasattr(model, 'childItems'):
        if parentType == 'folder':
            kwargs = kwargs.copy()
            kwargs['includeVirtual'] = True
        for item in model.childItems(
                parent, user=user,
                limit=_internal['limit'] + _internal['offset'],
                offset=0, sort=sort, **kwargs):
            if _internal['offset']:
                _internal['offset'] -= 1
            else:
                yield item
                if _internal['limit']:
                    _internal['limit'] -= 1
                    if not _internal['limit']:
                        _internal['done'] = True
                        return
    for folder in Folder().childFolders(
            parentType=parentType, parent=parent, user=user,
            limit=0, offset=0, sort=sort, **kwargs):
        if _internal['done']:
            return
        for item in allChildItems(folder, 'folder', user, sort=sort,
                                  _internal=_internal, **kwargs):
            yield item


@access.public(scope=TokenScope.DATA_READ)
@filtermodel(model=Item)
@autoDescribeRoute(
    Description('List items that match a query.')
    .responseClass('Item', array=True)
    .jsonParam('query', 'Find items that match this Mongo query.',
               required=True, requireObject=True)
    .pagingParams(defaultSort='_id')
    .errorResponse(),
)
@boundHandler()
def getItemsByQuery(self, query, limit, offset, sort):
    user = self.getCurrentUser()
    return Item().findWithPermissions(query, offset=offset, limit=limit, sort=sort, user=user)


@access.admin(scope=TokenScope.DATA_READ)
@filtermodel(model=File)
@autoDescribeRoute(
    Description('List files that match a query.')
    .responseClass('File', array=True)
    .jsonParam('query', 'Find files that match this Mongo query.',
               required=True, requireObject=True)
    .pagingParams(defaultSort='_id')
    .errorResponse(),
)
@boundHandler()
def getFilesByQuery(self, query, limit, offset, sort):
    user = self.getCurrentUser()
    return File().findWithPermissions(query, offset=offset, limit=limit, sort=sort, user=user)


@access.public(scope=TokenScope.DATA_READ)
@filtermodel(model=Folder)
@autoDescribeRoute(
    Description('List folders that match a query.')
    .responseClass('Folder', array=True)
    .jsonParam('query', 'Find folders that match this Mongo query.',
               required=True, requireObject=True)
    .pagingParams(defaultSort='_id')
    .errorResponse(),
)
@boundHandler()
def getFoldersByQuery(self, query, limit, offset, sort):
    user = self.getCurrentUser()
    return Folder().findWithPermissions(query, offset=offset, limit=limit, sort=sort, user=user)


@access.admin
@autoDescribeRoute(
    Description('Restart the Girder REST server.')
    .notes('Must be a system administrator to call this.')
    .errorResponse('You are not a system administrator.', 403),
)
@boundHandler()
def restartServer(self):
    import datetime

    import cherrypy
    from girder.utility import config

    if not config.getConfig()['server'].get('cherrypy_server', True):
        msg = 'Restarting of server is disabled.'
        raise RestException(msg, 403)

    class Restart(cherrypy.process.plugins.Monitor):
        def __init__(self, bus, frequency=1):
            cherrypy.process.plugins.Monitor.__init__(
                self, bus, self.run, frequency)

        def start(self):
            cherrypy.process.plugins.Monitor.start(self)

        def run(self):
            self.bus.log('Restarting.')
            self.thread.cancel()
            self.bus.restart()

    restart = Restart(cherrypy.engine)
    restart.subscribe()
    restart.start()
    return {'restarted': datetime.datetime.now(datetime.timezone.utc)}


@autoDescribeRoute(
    Description('Report on old jobs.')
    .param('age', 'The minimum age in days.', required=False,
           dataType='int', default=1)
    .param('status', 'A comma-separated list of statuses to include.  Blank '
           'for all.', required=False, default='0,1,2')
    .errorResponse(),
)
@access.admin
@boundHandler()
def getOldJobs(self, age, status):
    age = datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(-age)
    query = {'updated': {'$lt': age}}
    if status:
        query['status'] = {'$in': [int(s) for s in status.split(',')]}
    return Job().find(query, force=True).count()


@autoDescribeRoute(
    Description('Delete old jobs.')
    .param('age', 'The minimum age in days.', required=False,
           dataType='int', default=1)
    .param('status', 'A comma-separated list of statuses to include.  Blank '
           'for all.', required=False, default='0,1,2')
    .errorResponse(),
)
@access.admin
@boundHandler()
def deleteOldJobs(self, age, status):
    age = datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(-age)
    query = {'updated': {'$lt': age}}
    if status:
        query['status'] = {'$in': [int(s) for s in status.split(',')]}
    count = 0
    for job in Job().find(query, force=True):
        Job().remove(job)
        count += 1
    return count


class HUIResourceResource(ResourceResource):
    def __init__(self, apiRoot):
        super(ResourceResource, self).__init__()
        # Added to the resource route
        apiRoot.resource.route('GET', (':id', 'items'), self.getResourceItems)
        apiRoot.resource.route('PUT', ('metadata',), self.putResourceMetadata)

    @describeRoute(
        Description('Get all of the items that are children of a resource.')
        .param('id', 'The ID of the resource.', paramType='path')
        .param('type', 'The type of the resource (folder, collection, or '
               'user).')
        .pagingParams(defaultSort='_id')
        .errorResponse('ID was invalid.')
        .errorResponse('Access was denied for the resource.', 403),
    )
    @access.public(scope=TokenScope.DATA_READ)
    def getResourceItems(self, id, params):
        user = self.getCurrentUser()
        modelType = params['type']
        model = ModelImporter.model(modelType)
        doc = model.load(id=id, user=user, level=AccessType.READ)
        if not doc:
            msg = 'Resource not found.'
            raise RestException(msg)
        limit, offset, sort = self.getPagingParameters(params, '_id')
        return list(allChildItems(
            parentType=modelType, parent=doc, user=user,
            limit=limit, offset=offset, sort=sort))

    @autoDescribeRoute(
        Description('Set metadata on multiple resources at once.')
        .jsonParam('resources', 'A JSON-encoded set of resources to modify.  '
                   'Each type is a list of ids. For example: {"item": [(item '
                   'id 1), (item id 2)], "folder": [(folder id 1)]}.',
                   requireObject=True)
        .jsonParam('metadata', 'A JSON object containing the metadata keys to '
                   'add', paramType='body', requireObject=True)
        .param('allowNull', 'Whether "null" is allowed as a metadata value.',
               required=False, dataType='boolean', default=False)
        .errorResponse('Unsupported or unknown resource type.')
        .errorResponse('Invalid resources format.')
        .errorResponse('No resources specified.')
        .errorResponse('Resource not found.')
        .errorResponse('Write access was denied for a resource.', 403),
    )
    @access.public(scope=TokenScope.DATA_WRITE)
    def putResourceMetadata(self, resources, metadata, allowNull):
        user = self.getCurrentUser()
        self._validateResourceSet(resources)
        # Validate that we have write permission for all resources; if any
        # fail, no item will be changed.
        for kind in resources:
            model = self._getResourceModel(kind, 'setMetadata')
            for id in resources[kind]:
                model.load(id=id, user=user, level=AccessType.WRITE)
        metaUpdate = {}
        for key, value in metadata.items():
            if value is None and not allowNull:
                metaUpdate.setdefault('$unset', {})['meta.' + key] = ''
            else:
                metaUpdate.setdefault('$set', {})['meta.' + key] = value
        modified = 0
        for kind in resources:
            model = self._getResourceModel(kind, 'setMetadata')
            for id in resources[kind]:
                resource = model.load(id=id, user=user, level=AccessType.WRITE)
                # We aren't using model.setMetadata, since it is more
                # restrictive than we want.
                modified += model.update({'_id': resource['_id']}, metaUpdate).modified_count
        return modified


@access.admin(scope=TokenScope.SETTINGS_READ)
@autoDescribeRoute(
    Description('Get the value of a system setting, or a list of them.')
    .notes('Must be a system administrator to call this.')
    .param('key', 'The key identifying this setting.', required=False)
    .jsonParam('list', 'A JSON list of keys representing a set of settings to return.',
               required=False, requireArray=True)
    .param('default', 'If "none", return a null value if a setting is '
           'currently the default value. If "default", return the default '
           'value of the setting(s).', required=False)
    .errorResponse('You are not a system administrator.', 403),
)
@boundHandler
def getSettingDefault(self, key, list, default=None):
    getFunc = Setting().get
    if default == 'none':
        def getFuncValue(k):
            return (Setting()._get(k) or {}).get('value')
        getFunc = getFuncValue
    elif default == 'default':
        getFunc = Setting().getDefault
    elif default:
        msg = "Default was not 'none', 'default', or blank."
        raise RestException(msg)
    if list is not None:
        return {k: getFunc(k) for k in list}
    else:
        self.requireParams({'key': key})
        return getFunc(key)


@access.admin
@filtermodel(model=File)
@autoDescribeRoute(
    Description('Adjust the import path of a file.')
    .responseClass('File')
    .modelParam('id', model=File, level=AccessType.ADMIN)
    .param('path', 'The new import path of the file.')
    .errorResponse('ID was invalid.')
    .errorResponse('Write access was denied on the parent folder.', 403),

)
@boundHandler()
def adjustFileImportPath(self, file, path):
    assetstore = Assetstore().load(file['assetstoreId'])
    if assetstore['type'] != AssetstoreType.FILESYSTEM or not file.get('path'):
        msg = 'The file must be on a filesystem assetstore'
        raise RestException(msg)
    if not os.path.exists(path):
        msg = 'The new import path does not exist or is unreachable'
        raise RestException(msg)
    File().getAssetstoreAdapter(file).deleteFile(file)
    file['size'] = os.path.getsize(path)
    file['imported'] = True
    file['path'] = path
    file = File().save(file)
    try:
        import girder_hashsum_download as hashsum_download

        hashsum_download._computeHash(file)
    except ImportError:
        pass
    return file


@access.public(scope=TokenScope.DATA_READ)
@autoDescribeRoute(
    Description('Get resource paths for multiple resources at once.')
    .jsonParam('resources', 'A JSON-encoded set of resources to get resource paths.'
               'Each type is a list of ids. For example: {"item": [(item '
               'id 1), (item id 2)], "folder": [(folder id 1)]}.',
               paramType='body', requireObject=True),
)
@boundHandler()
def getMultipleResourcePaths(self, resources):
    user = self.getCurrentUser()
    results = {}
    for kind in resources:
        results[kind] = {}
        model = ModelImporter.model(kind)
        for id in resources[kind]:
            doc = model.load(id=id, user=user, level=AccessType.READ)
            if doc is None:
                logger.info(f'Failed to load {kind} {id}')
                continue
            try:
                results[kind][id] = path_util.getResourcePath(kind, doc, user=user)
            except Exception:
                logger.info(f'Failed to resolve path for {kind} {id} {doc}')
                continue
    return results



import os

import pooch

registry = {
    #
    # Source: Easy1.png
    'Easy1.png': 'sha512:feaf2b24c4ab3123caf1aa35f51acb1d8b83b34941b28d130f878702fd3be4ae9bf46176209f7511d1213da511d414c2b4b7738ad567b089224de9d6c189e664',  # noqa
    #
    # Source: sample_svs_image.TCGA-DU-6399-01A-01-TS1.e8eb65de-d63e-42db-af6f-14fefbbdf7bd.svs
    'sample_svs_image.TCGA-DU-6399-01A-01-TS1.e8eb65de-d63e-42db-af6f-14fefbbdf7bd.svs': 'sha512:5580c2b5a5360d279d102f1eb5b0e646a4943e362ec1d47f2db01f8e9e52b302e51692171198d0d35c7fa9ec9f5b8e445ef91fa7ea0bdb05ead31ab49e0118f9',  # noqa
}


class DKCPooch(pooch.Pooch):
    def get_url(self, fname):
        self._assert_file_in_registry(fname)
        algo, hashvalue = self.registry[fname].split(':')
        return self.base_url.format(algo=algo, hashvalue=hashvalue)


datastore = DKCPooch(
    path=pooch.utils.cache_location(
        os.path.join(os.environ.get('TOX_WORK_DIR', pooch.utils.os_cache('pooch')), 'externaldata'),
    ),
    base_url='https://data.kitware.com/api/v1/file/hashsum/{algo}/{hashvalue}/download',
    registry=registry,
    retry_if_failed=10,
)



from girder.models.folder import Folder
from girder.models.model_base import ValidationException

from .meta import TCGAModel


class Cohort(TCGAModel, Folder):

    TCGAType = 'cohort'

    def validate(self, doc, **kwargs):
        super().validate(doc, **kwargs)
        if doc['parentCollection'] != 'collection':
            msg = 'A Cohort model must be a child of a collection'
            raise ValidationException(
                msg,
            )
        return doc

    def importDocument(self, doc, **kwargs):
        from .case import Case

        recurse = kwargs.get('recurse', False)
        self.setTCGA(doc, cohort=doc['name'])
        doc = super().importDocument(
            doc, **kwargs)
        if not recurse:
            return doc

        childModel = Case()
        children = Folder().childFolders(
            doc, 'folder', user=kwargs.get('user'),
        )
        for child in children:
            try:
                childModel.importDocument(child, **kwargs)
            except ValidationException:
                pass
        return doc



from girder.api import access
from girder.api.describe import Description, describeRoute
from girder.api.rest import boundHandler, loadmodel
from girder.constants import AccessType
from girder.models.item import Item

from ..models.aperio import Aperio


def addItemEndpoints(itemRoot):
    itemRoot.route('POST', (':id', 'aperio'), importDocument)
    itemRoot.route('GET', (':id', 'aperio'), findAperio)
    itemRoot.route('DELETE', (':id', 'aperio'), removeAperio)
    itemRoot.route('PUT', (':id', 'aperio'), modifyAperio)


def addTcgaEndpoints(tcgaRoot):
    tcgaRoot.route('POST', ('aperio',), importTCGADocument)


@describeRoute(
    Description('Import an item as an Aperio annotation')
    .param('id', 'The id of the item or root to import')
    .param('tag', 'Import annotations with this tag',
           required=False)
    .param('recursive', 'Perform a recursive search for annotations',
           required=False, dataType='boolean'),
)
@access.admin
@loadmodel(model='item', level=AccessType.ADMIN)
@boundHandler()
def importTCGADocument(self, item, params):
    user = self.getCurrentUser()
    token = self.getCurrentToken()
    recursive = params.get('recursive')
    tag = params.get('tag')
    aperio = Aperio()
    return aperio.importTCGADocument(
        item, tag=tag,
        user=user, token=token,
        recurse=recursive,
    )


@describeRoute(
    Description('Import an item as an Aperio annotation')
    .param('id', 'The ID of the item containing the annotation file',
           paramType='path')
    .param('imageId', 'The ID of the slide image', required=True)
    .param('tag', 'A searchable tag to store with the metadata',
           required=False),
)
@access.admin
@loadmodel(model='item', level=AccessType.ADMIN)
@boundHandler()
def importDocument(self, item, params):
    self.requireParams('imageId', params)
    user = self.getCurrentUser()
    token = self.getCurrentToken()
    tag = params.get('tag')
    imageId = params['imageId']
    image = Item().load(
        imageId,
        user=user, level=AccessType.READ, exc=True,
    )
    aperio = Aperio()
    return aperio.importDocument(
        item, image, tag=tag,
        user=user, token=token,
    )


@describeRoute(
    Description('Find Aperio annotation items associated with a slide image.')
    .param('id', 'The ID of the slide image item', paramType='path')
    .param('tag', 'Filter by the given tag string', required=False)
    .pagingParams(defaultSort='name'),
)
@access.public
@loadmodel(model='item', level=AccessType.READ)
@boundHandler()
def findAperio(self, item, params):
    limit, offset, sort = self.getPagingParameters(params, 'name')
    user = self.getCurrentUser()
    tag = params.get('tag')
    aperio = Aperio()
    docs = list(aperio.findAperio(
        item, tag=tag,
        user=user, level=AccessType.READ,
    ))

    for doc in docs:
        files = aperio.childFiles(doc, limit=1)
        if files.count():
            doc['file'] = files[0]

    return docs


@describeRoute(
    Description('Remove Aperio specific metadata from an item')
    .param('id', 'The ID of the annotation item', paramType='path'),
)
@access.admin
@loadmodel(model='item', level=AccessType.WRITE)
@boundHandler()
def removeAperio(self, item, params):
    return Aperio().removeAperio(
        item,
    )


@describeRoute(
    Description('Set the tag associated with the annotation file')
    .param('id', 'The ID of the annotation file', paramType='path')
    .param('tag', 'A searchable tag to store with the metadata'),
)
@access.admin
@loadmodel(model='item', level=AccessType.WRITE)
@boundHandler()
def modifyAperio(self, item, params):
    return Aperio().setTag(
        item, tag=params.get('tag'),
    )



"""
Endpoints providing a simplified interface for handling TCGA datasets.
"""


import re

from girder import logger
from girder.api import access
from girder.api.describe import Description, describeRoute
from girder.api.rest import Resource, RestException, loadmodel
from girder.constants import AccessType, TokenScope
from girder.exceptions import ValidationException
from girder.models.collection import Collection
from girder.models.folder import Folder
from girder.models.item import Item
from girder.models.setting import Setting
from girder.models.user import User
from girder.utility import setting_utilities
from girder_jobs.constants import JobStatus
from girder_jobs.models.job import Job

from ..constants import TCGACollectionSettingKey
from ..models.aperio import Aperio
from ..models.case import Case
from ..models.cohort import Cohort
from ..models.image import Image
from ..models.pathology import Pathology
from ..models.slide import Slide

invalid_key_re = re.compile(r'[.$]')


def import_recursive(job):
    try:
        root = job['kwargs']['root']
        token = job['kwargs']['token']

        user = User().load(job['userId'], force=True)

        children = list(Folder().childFolders(
            root, 'collection', user=user,
        ))
        count = len(children)
        progress = 0

        job = Job().updateJob(
            job, log='Started TCGA import\n',
            status=JobStatus.RUNNING,
            progressCurrent=progress,
            progressTotal=count,
        )
        logger.info('Starting recursive TCGA import')

        for child in children:
            progress += 1
            try:
                msg = 'Importing "%s"' % child.get('name', '')
                job = Job().updateJob(
                    job, log=msg, progressMessage=msg + '\n',
                    progressCurrent=progress,
                )
                logger.debug(msg)
                Cohort().importDocument(
                    child, recurse=True, user=user, token=token, job=job,
                )
                job = Job().load(id=job['_id'], force=True)

                # handle any request to stop execution
                if (not job or job['status'] in (
                        JobStatus.CANCELED, JobStatus.ERROR)):
                    logger.info('TCGA import job halted with')
                    return

            except ValidationException:
                logger.warning('Failed to import %s' % child.get('name', ''))

        logger.info('Starting recursive TCGA import')
        job = Job().updateJob(
            job, log='Finished TCGA import\n',
            status=JobStatus.SUCCESS,
            progressCurrent=count,
            progressMessage='Finished TCGA import',
        )
    except Exception as e:
        logger.exception('Importing TCGA failed with %s' % str(e))
        job = Job().updateJob(
            job, log='Import failed with %s\n' % str(e),
            status=JobStatus.ERROR,
        )


def pagedResponse(cursor, limit, offset, sort):
    """Add paging headers to a paged query."""
    total = cursor.count()
    pages = (total - 1) // limit + 1
    index = offset // limit

    return {
        'total_count': total,
        'pos': offset,
        'limit': limit,
        'total_pages': pages,
        'current_page': index,
        'data': list(cursor),
    }


class TCGAResource(Resource):

    def __init__(self):
        super().__init__()

        @setting_utilities.validator({
            TCGACollectionSettingKey,
        })
        def validateTCGACollection(doc):
            model = Collection().load(
                doc['value'], force=True,
            )
            if model is None:
                msg = 'Invalid collection id'
                raise ValidationException(
                    msg, 'value',
                )

        self.resourceName = 'tcga'
        self.route('GET', (), self.getCollection)
        self.route('POST', (), self.setCollection)
        self.route('POST', ('import',), self.importCollection)
        self.route('DELETE', (), self.deleteCollection)

        self.route('GET', ('cohort',), self.findCohort)
        self.route('GET', ('cohort', ':id'), self.getCohort)
        self.route('POST', ('cohort',), self.importCohort)
        self.route('DELETE', ('cohort', ':id'), self.deleteCohort)
        self.route('GET', ('cohort', ':id', 'slides'), self.cohortListSlides)
        self.route('GET', ('cohort', ':id', 'images'), self.cohortListImages)

        self.route('GET', ('case',), self.findCase)
        self.route('GET', ('case', ':id'), self.getCase)
        self.route('GET', ('case', 'label', ':label'), self.getCaseByLabel)
        self.route('POST', ('case',), self.importCase)
        self.route('DELETE', ('case', ':id'), self.deleteCase)
        self.route('GET', ('case', 'search'), self.searchCase)
        self.route('GET', ('case', ':id', 'metadata', 'tables'), self.listCaseTables)
        self.route('GET', ('case', ':id', 'metadata', ':table'), self.getCaseMetadata)
        self.route('POST', ('case', ':id', 'metadata', ':table'), self.setCaseMetadata)
        self.route('PUT', ('case', ':id', 'metadata', ':table'), self.updateCaseMetadata)
        self.route('DELETE', ('case', ':id', 'metadata', ':table'), self.deleteCaseMetadata)
        self.route('GET', ('case', ':id', 'images'), self.listCaseImages)

        self.route('GET', ('slide',), self.findSlide)
        self.route('GET', ('slide', ':id'), self.getSlide)
        self.route('POST', ('slide',), self.importSlide)
        self.route('DELETE', ('slide', ':id'), self.deleteSlide)

        self.route('GET', ('image',), self.findImage)
        self.route('GET', ('image', ':id'), self.getImage)
        self.route('POST', ('image',), self.importImage)
        self.route('DELETE', ('image', ':id'), self.deleteImage)

        self.route('GET', ('pathology',), self.findPathology)
        self.route('GET', ('pathology', ':id'), self.getPathology)
        self.route('POST', ('pathology',), self.importPathology)
        self.route('DELETE', ('pathology', ':id'), self.deletePathology)

    def getTCGACollection(self, level=AccessType.READ):
        tcga = Setting().get(TCGACollectionSettingKey)
        if tcga is None:
            msg = 'TCGA collection id not initialized in settings'
            raise RestException(
                msg,
                code=404,
            )
        return Collection().load(
            tcga, level=level, user=self.getCurrentUser(),
        )

    @access.public(scope=TokenScope.DATA_READ)
    @describeRoute(
        Description('Get the TCGA collection'),
    )
    def getCollection(self, params):
        return self.getTCGACollection()

    @access.admin
    @describeRoute(
        Description('Set the TCGA collection')
        .param('collectionId', 'The id of the collection'),
    )
    def setCollection(self, params):
        user = self.getCurrentUser()
        self.requireParams('collectionId', params)

        # this is to ensure the collection exists
        collection = Collection().load(
            id=params['collectionId'], user=user,
            level=AccessType.WRITE, exc=True,
        )
        return Setting().set(
            TCGACollectionSettingKey,
            collection['_id'],
        )

    @access.admin
    @describeRoute(
        Description('Recursively import the TCGA collection')
        .notes('This will run as a local job that will execute '
               'asynchronously.'),
    )
    def importCollection(self, params):
        user = self.getCurrentUser()
        token = self.getCurrentToken()
        tcga = self.getTCGACollection(level=AccessType.WRITE)

        job = Job().createLocalJob(
            module='histomicsui.rest.tcga',
            function='import_recursive',
            kwargs={'root': tcga, 'token': token},
            title='Import TCGA items',
            user=user,
            type='tcga_import_recursive',
            public=False,
            asynchronous=True,
        )
        Job().scheduleJob(job)
        return job

    @access.admin
    @describeRoute(
        Description('Remove the TCGA collection'),
    )
    def deleteCollection(self, params):
        return Setting().unset(TCGACollectionSettingKey)

    # cohort endpoints
    #####################
    @access.public(scope=TokenScope.DATA_READ)
    @describeRoute(
        Description('List cohorts in the TCGA dataset')
        .pagingParams(defaultSort='name'),
    )
    def findCohort(self, params):
        user = self.getCurrentUser()
        tcga = self.getTCGACollection()
        limit, offset, sort = self.getPagingParameters(params, 'name')

        cursor = Cohort().childFolders(
            parentType='collection', parent=tcga, cursor=True,
            user=user, offset=offset, limit=limit, sort=sort,
        )
        return pagedResponse(cursor, limit, offset, sort)

    @access.public(scope=TokenScope.DATA_READ)
    @loadmodel(model='cohort', plugin='histomicsui',
               level=AccessType.READ)
    @describeRoute(
        Description('Get a cohort document from an id')
        .param('id', 'The id of the cohort', paramType='path'),
    )
    def getCohort(self, cohort, params):
        return cohort

    @access.admin
    @describeRoute(
        Description('Import a folder as a TCGA cohort type')
        .param('folderId', 'The id of the folder to import'),
    )
    def importCohort(self, params):
        user = self.getCurrentUser()
        token = self.getCurrentToken()
        self.requireParams('folderId', params)

        folder = Folder().load(
            id=params['folderId'], user=user,
            level=AccessType.WRITE, exc=True,
        )

        cohort = Cohort().importDocument(
            folder, user=user, token=token,
        )
        return cohort

    @access.admin
    @loadmodel(model='cohort', plugin='histomicsui',
               level=AccessType.WRITE)
    @describeRoute(
        Description('Remove a cohort type')
        .param('id', 'The id of the cohort', paramType='path'),
    )
    def deleteCohort(self, cohort, params):
        return Cohort().removeTCGA(
            cohort)

    @access.public
    @loadmodel(model='cohort', plugin='histomicsui',
               level=AccessType.READ)
    @describeRoute(
        Description('List slides in a cohort')
        .param('id', 'The id of the cohort', paramType='path')
        .pagingParams(defaultSort='name'),
    )
    def cohortListSlides(self, cohort, params):
        limit, offset, sort = self.getPagingParameters(params, 'name')
        cursor = Slide().find({
            'tcga.cohort': cohort['name'],
        }, cursor=True, limit=limit, offset=offset, sort=sort)
        return pagedResponse(cursor, limit, offset, sort)

    @access.public
    @loadmodel(model='cohort', plugin='histomicsui',
               level=AccessType.READ)
    @describeRoute(
        Description('List slide images in a cohort')
        .param('id', 'The id of the cohort', paramType='path')
        .pagingParams(defaultSort='name'),
    )
    def cohortListImages(self, cohort, params):
        limit, offset, sort = self.getPagingParameters(params, 'name')
        cursor = Image().find({
            'tcga.cohort': cohort['name'],
        }, cursor=True, limit=limit, offset=offset, sort=sort)
        return pagedResponse(cursor, limit, offset, sort)

    # Case endpoints
    #####################
    @access.public(scope=TokenScope.DATA_READ)
    @describeRoute(
        Description('List cases in the TCGA dataset')
        .param('cohort', 'The id of the cohort document', required=True)
        .pagingParams(defaultSort='name'),
    )
    def findCase(self, params):
        user = self.getCurrentUser()
        limit, offset, sort = self.getPagingParameters(params, 'name')
        cohort = Cohort().load(
            id=params['cohort'], user=user, level=AccessType.READ,
            exc=True,
        )

        cursor = Case().childFolders(
            parentType='folder', parent=cohort, cursor=True,
            user=user, offset=offset, limit=limit, sort=sort,
        )
        return pagedResponse(cursor, limit, offset, sort)

    @access.public(scope=TokenScope.DATA_READ)
    @loadmodel(model='case', plugin='histomicsui',
               level=AccessType.READ)
    @describeRoute(
        Description('Get a case document from an id')
        .param('id', 'The id of the case', paramType='path'),
    )
    def getCase(self, case, params):
        return case

    @access.public(scope=TokenScope.DATA_READ)
    @describeRoute(
        Description('Get a case document from a label')
        .param('label', 'The label of the case', paramType='path')
        .errorResponse('Label was invalid'),
    )
    def getCaseByLabel(self, label, params):
        user = self.getCurrentUser()
        case = Case().findOne(
            {'tcga.label': label}, user=user,
        )
        if not case:
            msg = 'TCGA case label not found'
            raise RestException(
                msg,
            )
        return case

    @access.admin
    @describeRoute(
        Description('Import a folder as a TCGA case')
        .param('folderId', 'The id of the folder to import'),
    )
    def importCase(self, params):
        user = self.getCurrentUser()
        token = self.getCurrentToken()
        self.requireParams('folderId', params)

        folder = Folder().load(
            id=params['folderId'], user=user,
            level=AccessType.WRITE, exc=True,
        )

        case = Case().importDocument(
            folder, user=user, token=token,
        )
        return case

    @access.admin
    @loadmodel(model='case', plugin='histomicsui',
               level=AccessType.WRITE)
    @describeRoute(
        Description('Remove a case document')
        .param('id', 'The id of the case', paramType='path'),
    )
    def deleteCase(self, case, params):
        return Case().removeTCGA(case)

    @access.public(scope=TokenScope.DATA_READ)
    @describeRoute(
        Description('Search for cases by clinical data')
        .param('table', 'A table to search',
               required=True)
        .param('key', 'A key that should be present',
               required=False)
        .param('value', 'The value associated with the given key',
               required=False)
        .param('substring', 'Find values containing this substring',
               required=False)
        .notes('Only one of "value" or "substring" may be provided.')
        .pagingParams(defaultSort='name'),
    )
    def searchCase(self, params):
        user = self.getCurrentUser()
        limit, offset, sort = self.getPagingParameters(params, 'name')
        self.requireParams('table', params)

        table = params.get('table')
        key = params.get('key')
        value = params.get('value')
        substring = params.get('substring')

        if value and substring:
            msg = 'Cannot search by both value and substring'
            raise RestException(
                msg,
            )

        if (value or substring) and not key:
            msg = 'A key must be provided to search by value'
            raise RestException(
                msg,
            )

        if key and invalid_key_re.search(key):
            msg = 'Invalid key parameter'
            raise RestException(
                msg,
            )

        query = {}
        if not key:
            query = {
                'tcga.meta.' + table: {
                    '$exists': True,
                },
            }
        elif not value and not substring:
            query = {
                'tcga.meta.' + table + '.' + key: {
                    '$exists': True,
                },
            }
        elif value:
            query = {
                'tcga.meta.' + table + '.' + key: value,
            }
        else:
            query = {
                'tcga.meta.' + table + '.' + key: re.compile(re.escape(substring)),
            }

        cursor = Case().find(
            query, user=user, offset=offset, limit=limit, sort=sort,
        )
        return pagedResponse(cursor, limit, offset, sort)

    @access.public(scope=TokenScope.DATA_READ)
    @loadmodel(model='case', plugin='histomicsui',
               level=AccessType.READ)
    @describeRoute(
        Description('List tables present inside case metadata')
        .param('id', 'The id of the case', paramType='path'),
    )
    def listCaseTables(self, case, params):
        return list(Case().getTCGAMeta(
            case).keys())

    @access.public(scope=TokenScope.DATA_READ)
    @loadmodel(model='case', plugin='histomicsui',
               level=AccessType.READ)
    @describeRoute(
        Description('Get case metadata')
        .param('id', 'The id of the case', paramType='path')
        .param('table', 'The table name to get', paramType='path'),
    )
    def getCaseMetadata(self, case, table, params):
        return Case().getTCGAMeta(
            case).get(table)

    @access.user(scope=TokenScope.DATA_WRITE)
    @loadmodel(model='case', plugin='histomicsui',
               level=AccessType.WRITE)
    @describeRoute(
        Description('Create or replace case metadata')
        .param('id', 'The id of the case', paramType='path')
        .param('table', 'The table to update', paramType='path')
        .param('body', 'A JSON object containing the metadata to create',
               paramType='body'),
    )
    def setCaseMetadata(self, case, table, params):
        metadata = self.getBodyJson()
        caseModel = Case()
        for k in metadata:
            if not len(k) or '.' in k or k[0] == '$':
                msg = 'Invalid key name'
                raise RestException(
                    msg,
                )
        meta = caseModel.getTCGAMeta(case)
        meta[table] = metadata
        caseModel.save(case)
        return metadata

    @access.user(scope=TokenScope.DATA_WRITE)
    @loadmodel(model='case', plugin='histomicsui',
               level=AccessType.WRITE)
    @describeRoute(
        Description('Update case metadata')
        .notes('Set metadata fields to null to delete them.')
        .param('id', 'The id of the case', paramType='path')
        .param('table', 'The table to update', paramType='path')
        .param('body', 'A JSON object containing the metadata to update',
               paramType='body'),
    )
    def updateCaseMetadata(self, case, table, params):
        metadata = self.getBodyJson()
        caseModel = Case()
        for k in metadata:
            if not len(k) or '.' in k or k[0] == '$':
                msg = 'Invalid key name'
                raise RestException(
                    msg,
                )
        meta = {
            table: metadata,
        }

        caseModel.updateTCGAMeta(case, meta).save(case)
        return caseModel.getTCGAMeta(case).get(table)

    @access.user(scope=TokenScope.DATA_WRITE)
    @loadmodel(model='case', plugin='histomicsui',
               level=AccessType.WRITE)
    @describeRoute(
        Description('Delete case metadata')
        .param('id', 'The id of the case', paramType='path')
        .param('table', 'The table to remove', paramType='path'),
    )
    def deleteCaseMetadata(self, case, table, params):
        caseModel = Case()
        meta = caseModel.getTCGAMeta(case)
        del meta[table]
        caseModel.save(case)

    @access.public
    @loadmodel(model='case', plugin='histomicsui',
               level=AccessType.READ)
    @describeRoute(
        Description('List images under a case')
        .param('id', 'The id of the case', paramType='path')
        .pagingParams(defaultSort='name'),
    )
    @access.public(scope=TokenScope.DATA_READ)
    def listCaseImages(self, case, params):
        limit, offset, sort = self.getPagingParameters(params, 'name')
        cursor = Image().find({
            'tcga.label': case['name'],
        }, cursor=True, limit=limit, offset=offset, sort=sort)
        return pagedResponse(cursor, limit, offset, sort)

    # Slide endpoints
    #####################
    @access.public(scope=TokenScope.DATA_READ)
    @describeRoute(
        Description('Find slides for a case')
        .param('case', 'The id of case document', required=True)
        .pagingParams(defaultSort='name'),
    )
    def findSlide(self, params):
        limit, offset, sort = self.getPagingParameters(params, 'name')
        user = self.getCurrentUser()
        case = Case().load(
            id=params['case'], user=user, level=AccessType.READ,
            exc=True,
        )
        cursor = Slide().childFolders(
            parentType='folder', parent=case, user=user, cursor=True,
            offset=offset, limit=limit, sort=sort,
        )
        return pagedResponse(cursor, limit, offset, sort)

    @access.public(scope=TokenScope.DATA_READ)
    @loadmodel(model='slide', plugin='histomicsui',
               level=AccessType.READ)
    @describeRoute(
        Description('Get a slide document by id')
        .param('id', 'The id of the slide', paramType='path'),
    )
    def getSlide(self, slide, params):
        return slide

    @access.admin
    @describeRoute(
        Description('Import a folder as a TCGA slide')
        .param('folderId', 'The id of the folder to import'),
    )
    def importSlide(self, params):
        user = self.getCurrentUser()
        token = self.getCurrentToken()
        self.requireParams('folderId', params)

        folder = Folder().load(
            id=params['folderId'], user=user,
            level=AccessType.WRITE, exc=True,
        )

        slide = Slide().importDocument(
            folder, user=user, token=token,
        )
        return slide

    @access.admin
    @loadmodel(model='slide', plugin='histomicsui',
               level=AccessType.WRITE)
    @describeRoute(
        Description('Remove a slide')
        .param('id', 'The id of the slide', paramType='path'),
    )
    def deleteSlide(self, slide, params):
        return Slide().removeTCGA(slide)

    # Image endpoints
    #####################
    @access.public(scope=TokenScope.DATA_READ)
    @describeRoute(
        Description('Find images')
        .param('slide', 'The id of slide document', required=False)
        .param('caseName', 'The name of the case', required=False)
        .pagingParams(defaultSort='name'),
    )
    def findImage(self, params):
        limit, offset, sort = self.getPagingParameters(params, 'name')
        user = self.getCurrentUser()
        if params.get('slide'):
            slide = Slide().load(
                id=params['slide'], user=user, level=AccessType.READ,
                exc=True,
            )
            cursor = Image().find(
                {'folderId': slide['_id']},
                offset=offset, limit=limit, sort=sort,
            )

        elif params.get('caseName'):
            cursor = Image().find({
                'tcga.label': params['caseName'],
            }, cursor=True, limit=limit, offset=offset, sort=sort)

        else:
            msg = 'You must provide a slide id or case name'
            raise RestException(msg)
        return pagedResponse(cursor, limit, offset, sort)

    @access.public(scope=TokenScope.DATA_READ)
    @loadmodel(model='image', plugin='histomicsui',
               level=AccessType.READ)
    @describeRoute(
        Description('Get an image document by id')
        .param('id', 'The id of the image', paramType='path'),
    )
    def getImage(self, image, params):
        return image

    @access.admin
    @describeRoute(
        Description('Import an item as a TCGA slide image')
        .param('itemId', 'The id of the item to import'),
    )
    def importImage(self, params):
        user = self.getCurrentUser()
        token = self.getCurrentToken()
        self.requireParams('itemId', params)

        item = Item().load(
            id=params['itemId'], user=user,
            level=AccessType.WRITE, exc=True,
        )

        image = Image().importDocument(
            item, user=user, token=token,
        )
        return image

    @access.admin
    @loadmodel(model='image', plugin='histomicsui',
               level=AccessType.WRITE)
    @describeRoute(
        Description('Remove an image')
        .param('id', 'The id of the image', paramType='path'),
    )
    def deleteImage(self, image, params):
        return Image().removeTCGA(image)

    # Pathology endpoints
    #####################
    @access.public(scope=TokenScope.DATA_READ)
    @describeRoute(
        Description('Find pathologies for a case')
        .param('case', 'The id of a case document', required=True)
        .pagingParams(defaultSort='name'),
    )
    def findPathology(self, params):
        limit, offset, sort = self.getPagingParameters(params, 'name')
        user = self.getCurrentUser()
        case = Case().load(
            id=params['case'], user=user, level=AccessType.READ,
            exc=True,
        )
        model = Pathology()
        cursor = model.find(
            {'tcga.case': case['tcga']['label']},
            offset=offset, limit=limit, sort=sort,
        )

        # Inject file information into the returned documents
        response = pagedResponse(cursor, limit, offset, sort)
        for doc in response.get('data', []):
            files = model.childFiles(doc, limit=1)
            if files.count():
                doc['file'] = files[0]
        return response

    @access.public(scope=TokenScope.DATA_READ)
    @loadmodel(model='pathology', plugin='histomicsui',
               level=AccessType.READ)
    @describeRoute(
        Description('Get a pathology document by id')
        .param('id', 'The id of the pathology', paramType='path'),
    )
    def getPathology(self, pathology, params):
        files = Pathology().childFiles(
            pathology, limit=1,
        )
        if files.count():
            pathology['file'] = files[0]
        return pathology

    @access.admin
    @describeRoute(
        Description('Import an item as a TCGA pathology')
        .param('id', 'The id of the item to import')
        .param('recursive', 'Perform a recursive search for pathologies',
               required=False, dataType='boolean'),
    )
    def importPathology(self, params):
        user = self.getCurrentUser()
        token = self.getCurrentToken()
        self.requireParams('id', params)

        item = Pathology().loadDocument(
            id=params['id'], user=user,
            level=AccessType.WRITE, exc=True,
        )

        pathology = Pathology().importDocument(
            item, user=user, token=token, recurse=params.get('recursive'),
        )
        return pathology

    @access.admin
    @loadmodel(model='pathology', plugin='histomicsui',
               level=AccessType.WRITE)
    @describeRoute(
        Description('Remove a pathology')
        .param('id', 'The id of the pathology', paramType='path'),
    )
    def deletePathology(self, pathology, params):
        return Pathology().removeTCGA(pathology)

    # Aperio endpoints
    #####################
    @access.public(scope=TokenScope.DATA_READ)
    @describeRoute(
        Description('Find Aperio annotations for a case')
        .param('case', 'The id of a case document', required=True)
        .pagingParams(defaultSort='name'),
    )
    def findAperio(self, params):
        limit, offset, sort = self.getPagingParameters(params, 'name')
        user = self.getCurrentUser()
        case = Case().load(
            id=params['case'], user=user, level=AccessType.READ,
            exc=True,
        )
        cursor = Aperio().find(
            {'tcga.case': case['tcga']['label']},
            offset=offset, limit=limit, sort=sort,
        )
        return pagedResponse(cursor, limit, offset, sort)

    @access.public(scope=TokenScope.DATA_READ)
    @loadmodel(model='aperio', plugin='histomicsui',
               level=AccessType.READ)
    @describeRoute(
        Description('Get an Aperio document by id')
        .param('id', 'The id of the Aperio item', paramType='path'),
    )
    def getAperio(self, aperio, params):
        return aperio

    @access.admin
    @describeRoute(
        Description('Import an item as a TCGA Aperio XML item')
        .param('id', 'The id of the item or root to import')
        .param('recursive', 'Perform a recursive search for annotations',
               required=False, dataType='boolean'),
    )
    def importAperio(self, params):
        user = self.getCurrentUser()
        token = self.getCurrentToken()
        self.requireParams('id', params)

        item = Aperio().loadDocument(
            id=params['id'], user=user,
            level=AccessType.WRITE, exc=True,
        )

        aperio = Aperio().importDocument(
            item, user=user, token=token, recurse=params.get('recursive'),
        )
        return aperio

    @access.admin
    @loadmodel(model='aperio', plugin='histomicsui',
               level=AccessType.WRITE)
    @describeRoute(
        Description('Remove an Aperio XML item')
        .param('id', 'The id of the Aperio item', paramType='path'),
    )
    def deleteAperio(self, aperio, params):
        return Aperio().removeTCGA(aperio)



import pytest
from girder.models.folder import Folder
from girder.models.item import Item

from . import girder_utilities as utilities


@pytest.mark.plugin('histomicsui')
class TestImageBrowseEndpoints:
    def makeResources(self, admin):
        self.folder = list(Folder().childFolders(admin, 'user', user=admin))[0]
        self.items = [
            Item().createItem('item_%i' % i, creator=admin, folder=self.folder)
            for i in range(10)
        ]
        for item in self.items:
            # make the item look like an image
            item['largeImage'] = {
                'fileId': 'deadbeef',
            }
            Item().save(item)
        self.nonimage = Item().createItem('non-image', creator=admin, folder=self.folder)
        self.extraFolder = Folder().createFolder(self.folder, 'extra', creator=admin)

    def testGetNextImage(self, server, admin):
        self.makeResources(admin)
        resp = server.request(
            path='/item/%s/next_image' % str(self.items[0]['_id']), user=admin)
        assert utilities.respStatus(resp) == 200
        assert resp.json['_id'] == str(self.items[1]['_id'])

        resp = server.request(
            path='/item/%s/next_image' % str(self.items[-1]['_id']), user=admin)
        assert utilities.respStatus(resp) == 200
        assert resp.json['_id'] == str(self.items[0]['_id'])

    def testGetPreviousImage(self, server, admin):
        self.makeResources(admin)
        resp = server.request(
            path='/item/%s/previous_image' % str(self.items[0]['_id']), user=admin)
        assert utilities.respStatus(resp) == 200
        assert resp.json['_id'] == str(self.items[-1]['_id'])

        resp = server.request(
            path='/item/%s/previous_image' % str(self.items[-1]['_id']), user=admin)
        assert utilities.respStatus(resp) == 200
        assert resp.json['_id'] == str(self.items[-2]['_id'])

    def testGetNextImageException(self, server, admin):
        self.makeResources(admin)
        resp = server.request(
            path='/item/%s/next_image' % str(self.nonimage['_id']), user=admin)
        assert utilities.respStatus(resp) == 404

    def testGetNextImageFolder(self, server, admin):
        self.makeResources(admin)
        folderId = str(self.folder['_id'])
        resp = server.request(
            path='/item/%s/next_image' % str(self.items[0]['_id']),
            user=admin, params=dict(folderId=folderId))
        assert utilities.respStatus(resp) == 200
        assert resp.json['_id'] == str(self.items[1]['_id'])

        resp = server.request(
            path='/item/%s/next_image' % str(self.items[-1]['_id']),
            user=admin, params=dict(folderId=folderId))
        assert utilities.respStatus(resp) == 200
        assert resp.json['_id'] == str(self.items[0]['_id'])

    def testGetPreviousImageFolder(self, server, admin):
        self.makeResources(admin)
        folderId = str(self.folder['_id'])
        resp = server.request(
            path='/item/%s/previous_image' % str(self.items[0]['_id']),
            user=admin, params=dict(folderId=folderId))
        assert utilities.respStatus(resp) == 200
        assert resp.json['_id'] == str(self.items[-1]['_id'])

        resp = server.request(
            path='/item/%s/previous_image' % str(self.items[-1]['_id']),
            user=admin, params=dict(folderId=folderId))
        assert utilities.respStatus(resp) == 200
        assert resp.json['_id'] == str(self.items[-2]['_id'])

    def testGetNextImageExceptionFolder(self, server, admin):
        self.makeResources(admin)
        resp = server.request(
            path='/item/%s/next_image' % str(self.items[0]['_id']),
            user=admin, params=dict(folderId=self.extraFolder['_id']))
        assert utilities.respStatus(resp) == 404

    def testGetAdjacentImages(self, server, admin):
        self.makeResources(admin)
        resp = server.request(
            path='/item/%s/adjacent_images' % str(self.items[0]['_id']), user=admin)
        assert utilities.respStatus(resp) == 200
        assert resp.json['next']['_id'] == str(self.items[1]['_id'])

        resp = server.request(
            path='/item/%s/adjacent_images' % str(self.items[-1]['_id']), user=admin)
        assert utilities.respStatus(resp) == 200
        assert resp.json['next']['_id'] == str(self.items[0]['_id'])



"""Test histomicsui endpoints"""

import json

import pytest
from girder.exceptions import ValidationException
from girder.models.collection import Collection
from girder.models.folder import Folder
from girder.models.group import Group
from girder.models.item import Item
from girder.models.setting import Setting
from girder.models.user import User
from girder.utility import config

from histomicsui.constants import PluginSettings

from . import girder_utilities as utilities

# This probably should be moved into a fixture
config.getConfig()['histomicsui'] = {'restrict_downloads': True}


@pytest.mark.plugin('histomicsui')
class TestHUIResourceAndItem:
    def makeResources(self, admin):
        self.publicFolder = utilities.namedFolder(admin, 'Public')
        self.privateFolder = utilities.namedFolder(admin, 'Private')
        # Create some resources to use in the tests
        self.collection = Collection().createCollection(
            'collection A', admin)
        self.colFolderA = Folder().createFolder(
            self.collection, 'folder A', parentType='collection',
            creator=admin)
        self.colFolderB = Folder().createFolder(
            self.collection, 'folder B', parentType='collection',
            creator=admin)
        self.colFolderC = Folder().createFolder(
            self.colFolderA, 'folder C', creator=admin)
        self.colItemA1 = Item().createItem('item A1', admin, self.colFolderA)
        self.colItemB1 = Item().createItem('item B1', admin, self.colFolderB)
        self.colItemB2 = Item().createItem('item B2', admin, self.colFolderB)
        self.colItemC1 = Item().createItem('item C1', admin, self.colFolderC)
        self.colItemC2 = Item().createItem('item C2', admin, self.colFolderC)
        self.colItemC3 = Item().createItem('item C3', admin, self.colFolderC)
        self.itemPub1 = Item().createItem('item Public 1', admin, self.publicFolder)
        self.itemPriv1 = Item().createItem('item Private 1', admin, self.privateFolder)
        self.folderD = Folder().createFolder(self.publicFolder, 'folder D', creator=admin)
        self.itemD1 = Item().createItem('item D1', admin, self.folderD)
        self.itemD2 = Item().createItem('item D2', admin, self.folderD)

    def testResourceItems(self, server, admin):
        self.makeResources(admin)
        # Now test that we get the items we expect
        # From a user
        resp = server.request(
            path='/resource/%s/items' % admin['_id'], user=admin,
            params={'type': 'user'})
        assert utilities.respStatus(resp) == 200
        items = resp.json
        assert [item['name'] for item in items] == [
            'item Public 1', 'item D1', 'item D2', 'item Private 1']
        # From a collection
        resp = server.request(
            path='/resource/%s/items' % self.collection['_id'], user=admin,
            params={'type': 'collection'})
        assert utilities.respStatus(resp) == 200
        items = resp.json
        assert [item['name'] for item in items] == [
            'item A1', 'item C1', 'item C2', 'item C3', 'item B1', 'item B2']
        # With sort, limit, and offset
        resp = server.request(
            path='/resource/%s/items' % self.collection['_id'], user=admin,
            params={'type': 'collection', 'limit': 4, 'offset': 1,
                    'sort': 'name', 'sortdir': -1})
        assert utilities.respStatus(resp) == 200
        items = resp.json
        assert [item['name'] for item in items] == [
            'item B1', 'item A1', 'item C3', 'item C2']
        resp = server.request(
            path='/resource/%s/items' % self.collection['_id'], user=admin,
            params={'type': 'collection', 'limit': 1, 'offset': 0,
                    'sort': 'name', 'sortdir': -1})
        assert utilities.respStatus(resp) == 200
        items = resp.json
        assert [item['name'] for item in items] == ['item B2']
        # From a folder
        resp = server.request(
            path='/resource/%s/items' % self.colFolderA['_id'],
            user=admin, params={'type': 'folder'})
        assert utilities.respStatus(resp) == 200
        items = resp.json
        assert [item['name'] for item in items] == [
            'item A1', 'item C1', 'item C2', 'item C3']
        # From a lower folder
        resp = server.request(
            path='/resource/%s/items' % self.colFolderC['_id'],
            user=admin, params={'type': 'folder'})
        assert utilities.respStatus(resp) == 200
        items = resp.json
        assert [item['name'] for item in items] == ['item C1', 'item C2', 'item C3']

        # With a bad parameter
        resp = server.request(
            path='/resource/%s/items' % self.colFolderC['_id'],
            user=admin, params={'type': 'collection'})
        assert utilities.respStatus(resp) == 400
        assert 'Resource not found' in resp.json['message']

    def testItemQuery(self, server, admin, user):
        self.makeResources(admin)
        itemMeta = [
            {'key1': 'value1'},
            {'key1': 'value2'},
            {'key1': 'value1', 'key2': 'value2'},
        ]
        for idx, meta in enumerate(itemMeta):
            item = Item().createItem('item %d' % idx, admin, self.privateFolder)
            item['meta'] = meta
            item = Item().save(item)
        resp = server.request(
            path='/item/query', user=admin, params={'query': json.dumps({
                'meta.key1': {'$exists': True},
            })})
        assert utilities.respStatus(resp) == 200
        items = resp.json
        assert len(items) == 3
        resp = server.request(
            path='/item/query', user=user, params={'query': json.dumps({
                'meta.key1': {'$exists': True},
            })})
        assert utilities.respStatus(resp) == 200
        items = resp.json
        assert len(items) == 0
        resp = server.request(
            path='/item/query', user=admin, params={'query': json.dumps({
                'meta.key1': 'value1',
            })})
        assert utilities.respStatus(resp) == 200
        items = resp.json
        assert len(items) == 2
        resp = server.request(
            path='/item/query', user=admin, params={'query': json.dumps({
                'meta': {'key1': 'value1'},
            })})
        assert utilities.respStatus(resp) == 200
        items = resp.json
        assert len(items) == 1

    def testFolderQuery(self, server, admin, user):
        self.makeResources(admin)
        resp = server.request(
            path='/folder/query', user=admin, params={'query': json.dumps({
                'baseParentId': {'$oid': str(admin['_id'])},
            })})
        assert utilities.respStatus(resp) == 200
        assert len(resp.json) == 3
        resp = server.request(
            path='/folder/query', user=user, params={'query': json.dumps({
                'baseParentId': {'$oid': str(admin['_id'])},
            })})
        assert utilities.respStatus(resp) == 200
        assert len(resp.json) == 2

    def testResourceMetadata(self, server, admin):
        self.makeResources(admin)
        resp = server.request(
            method='PUT', path='/resource/metadata', params={
                'resources': json.dumps({'item': [str(self.colItemA1['_id'])]}),
                'metadata': json.dumps({
                    'keya': 'valuea',
                    'keyb.keyc': 'valuec',
                })})
        assert utilities.respStatus(resp) == 401
        resp = server.request(
            method='PUT', path='/resource/metadata', user=admin, params={
                'resources': json.dumps({'item': [str(self.colItemA1['_id'])]}),
                'metadata': json.dumps({
                    'keya': 'valuea',
                    'keyb.keyc': 'valuec',
                })})
        assert utilities.respStatus(resp) == 200
        assert resp.json == 1
        meta = Item().load(self.colItemA1['_id'], user=admin)['meta']
        assert meta['keya'] == 'valuea'
        assert meta['keyb']['keyc'] == 'valuec'
        resp = server.request(
            method='PUT', path='/resource/metadata', user=admin, params={
                'resources': json.dumps({'item': [
                    str(self.colItemA1['_id']),
                    str(self.colItemB1['_id']),
                ], 'folder': [str(self.colFolderA['_id'])]}),
                'metadata': json.dumps({
                    'keya': 'valuea',
                    'keyb.keyc': None,
                    'keyb.keyd': 'valued',
                })})
        assert utilities.respStatus(resp) == 200
        assert resp.json == 3
        meta = Item().load(self.colItemA1['_id'], user=admin)['meta']
        assert meta['keya'] == 'valuea'
        assert 'keyc' not in meta['keyb']
        assert meta['keyb']['keyd'] == 'valued'
        resp = server.request(
            method='PUT', path='/resource/metadata', user=admin, params={
                'resources': json.dumps({'item': [
                    str(self.colItemA1['_id']),
                ], 'folder': [str(self.colFolderA['_id'])]}),
                'metadata': json.dumps({
                    'keya': 'valuea',
                    'keyb.keyc': None,
                    'keyb.keyd': 'valued',
                }),
                'allowNull': True})
        assert utilities.respStatus(resp) == 200
        assert resp.json == 2
        meta = Item().load(self.colItemA1['_id'], user=admin)['meta']
        assert meta['keya'] == 'valuea'
        assert meta['keyb']['keyc'] is None
        assert meta['keyb']['keyd'] == 'valued'


@pytest.mark.plugin('histomicsui')
class TestHUIEndpoints:
    def makeResources(self, admin):
        self.user2 = User().createUser(
            email='user2@email.com', login='user2', firstName='user2',
            lastName='user2', password='password', admin=False)
        self.group = self.group = Group().createGroup('test group', creator=self.user2)
        Group().addUser(self.group, self.user2)

    def testHUISettings(self, server):
        for key in {
                PluginSettings.HUI_DEFAULT_DRAW_STYLES,
                PluginSettings.HUI_PANEL_LAYOUT}:

            resp = server.request(path='/histomicsui/settings')
            assert utilities.respStatus(resp) == 200
            settings = resp.json
            assert settings[key] is None
            assert settings[PluginSettings.HUI_BRAND_NAME] == 'HistomicsUI'

            Setting().set(key, '')
            assert Setting().get(key) is None
            with pytest.raises(ValidationException, match='must be a JSON'):
                Setting().set(key, 'not valid')
            with pytest.raises(ValidationException, match='must be a JSON'):
                Setting().set(key, json.dumps({'not': 'a list'}))
            value = [{'lineWidth': 8, 'id': 'Group 8'}]
            Setting().set(key, json.dumps(value))
            assert json.loads(Setting().get(key)) == value
            Setting().set(key, value)
            assert json.loads(Setting().get(key)) == value

            resp = server.request(path='/histomicsui/settings')
            assert utilities.respStatus(resp) == 200
            settings = resp.json
            assert json.loads(settings[key]) == value

    def testGeneralSettings(self, server, admin, user):
        self.makeResources(admin)
        settings = [{
            'key': PluginSettings.HUI_WEBROOT_PATH,
            'initial': 'histomics',
            'bad': {
                'girder': 'not be "girder"',
                '': 'not be empty',
            },
            'good': {'alternate1': 'alternate1'},
        }, {
            'key': PluginSettings.HUI_BRAND_NAME,
            'initial': 'HistomicsUI',
            'bad': {'': 'not be empty'},
            'good': {'Alternate': 'Alternate'},
        }, {
            'key': PluginSettings.HUI_BRAND_COLOR,
            'initial': '#777777',
            'bad': {
                '': 'not be empty',
                'white': 'be a hex color',
                '#777': 'be a hex color',
            },
            'good': {'#000000': '#000000'},
        }, {
            'key': PluginSettings.HUI_BANNER_COLOR,
            'initial': '#f8f8f8',
            'bad': {
                '': 'not be empty',
                'white': 'be a hex color',
                '#777': 'be a hex color',
            },
            'good': {'#000000': '#000000'},
        }, {
            'key': PluginSettings.HUI_ALTERNATE_WEBROOT_PATH,
            'initial': None,
            'bad': {
                'girder': 'not contain "girder"',
                'girder,histomicstk': 'not contain "girder"',
            },
            'good': {
                '': '',
                'histomicstk': 'histomicstk',
                'hui,histomicstk': 'hui,histomicstk',
            },
        }]
        for setting in settings:
            key = setting['key']
            assert Setting().get(key) == setting['initial']
            for badval in setting.get('bad', {}):
                with pytest.raises(ValidationException, match=setting['bad'][badval]):
                    Setting().set(key, badval)
            for badval in setting.get('badjson', []):
                with pytest.raises(ValidationException, match=badval['return']):
                    Setting().set(key, badval['value'])
            for goodval in setting.get('good', {}):
                assert Setting().set(key, goodval)['value'] == setting['good'][goodval]
            for goodval in setting.get('goodjson', []):
                assert Setting().set(key, goodval['value'])['value'] == goodval['return']

    def testGetWebroot(self, server):
        resp = server.request(path='/histomics', method='GET', isJson=False, prefix='')
        assert utilities.respStatus(resp) == 200
        body = utilities.getBody(resp)
        assert '<title>HistomicsUI</title>' in body
        resp = server.request(path='/alternate2', method='GET', isJson=False, prefix='')
        assert utilities.respStatus(resp) == 404
        Setting().set(PluginSettings.HUI_WEBROOT_PATH, 'alternate2')
        Setting().set(PluginSettings.HUI_BRAND_NAME, 'Alternate')
        resp = server.request(path='/histomics', method='GET', isJson=False, prefix='')
        assert utilities.respStatus(resp) == 200
        body = utilities.getBody(resp)
        assert '<title>Alternate</title>' in body
        resp = server.request(path='/alternate2', method='GET', isJson=False, prefix='')
        assert utilities.respStatus(resp) == 200
        body = utilities.getBody(resp)
        assert '<title>Alternate</title>' in body

    def testRestrictDownloads(self, server, fsAssetstore, admin, user):
        self.makeResources(admin)
        file = utilities.uploadExternalFile('Easy1.png', admin, fsAssetstore)
        resp = server.request(
            path='/item/%s/download' % file['itemId'], user=self.user2, isJson=False)
        assert utilities.respStatus(resp) == 200
        resp = server.request(
            path='/item/%s/download' % file['itemId'], user=None)
        assert utilities.respStatus(resp) == 401
        resp = server.request(
            path='/item/%s/tiles/images/noimage' % file['itemId'], user=self.user2)
        assert utilities.respStatus(resp) == 400
        resp = server.request(
            path='/item/%s/tiles/images/noimage' % file['itemId'], user=None)
        assert utilities.respStatus(resp) == 401

    def testQuarantine(self, server, admin, user):
        publicFolder = Folder().childFolders(  # noqa: B305
            user, 'user', filters={'name': 'Public'},
        ).next()
        adminFolder = Folder().childFolders(  # noqa: B305
            admin, 'user', filters={'name': 'Public'},
        ).next()
        privateFolder = Folder().childFolders(  # noqa: B305
            admin, 'user', filters={'name': 'Private'}, user=admin,
        ).next()
        items = [Item().createItem(name, creator, folder) for name, creator, folder in [
            ('userPublic1', user, publicFolder),
            ('userPublic2', user, publicFolder),
            ('adminPublic1', admin, adminFolder),
            ('adminPublic2', admin, adminFolder),
            ('adminPrivate1', admin, privateFolder),
            ('adminPrivate2', admin, privateFolder),
        ]]
        resp = server.request(
            method='PUT',
            path='/histomicsui/quarantine/%s' % str(items[0]['_id']))
        assert utilities.respStatus(resp) == 401
        assert 'Write access denied' in resp.json['message']
        resp = server.request(
            method='PUT', user=user,
            path='/histomicsui/quarantine/%s' % str(items[0]['_id']))
        assert utilities.respStatus(resp) == 400
        assert 'The quarantine folder is not configure' in resp.json['message']
        key = PluginSettings.HUI_QUARANTINE_FOLDER
        Setting().set(key, str(privateFolder['_id']))
        resp = server.request(
            method='PUT', user=user,
            path='/histomicsui/quarantine/%s' % str(items[0]['_id']))
        assert utilities.respStatus(resp) == 200
        resp = server.request(
            method='PUT', user=user,
            path='/histomicsui/quarantine/%s' % str(items[0]['_id']))
        assert utilities.respStatus(resp) == 403
        assert 'Write access denied' in resp.json['message']
        resp = server.request(
            method='PUT', user=user,
            path='/histomicsui/quarantine/%s' % str(items[2]['_id']))
        assert utilities.respStatus(resp) == 403
        assert 'Write access denied' in resp.json['message']
        resp = server.request(
            method='PUT', user=admin,
            path='/histomicsui/quarantine/%s' % str(items[2]['_id']))
        assert utilities.respStatus(resp) == 200
        resp = server.request(
            method='PUT', user=admin,
            path='/histomicsui/quarantine/%s' % str(items[4]['_id']))
        assert utilities.respStatus(resp) == 400
        assert 'already in the quarantine' in resp.json['message']
        # Restore
        resp = server.request(
            method='PUT', user=admin,
            path='/histomicsui/quarantine/%s/restore' % str(items[1]['_id']))
        assert utilities.respStatus(resp) == 400
        assert 'no quarantine record' in resp.json['message']
        resp = server.request(
            method='PUT',
            path='/histomicsui/quarantine/%s/restore' % str(items[0]['_id']))
        assert utilities.respStatus(resp) == 401
        assert 'Write access denied' in resp.json['message']
        resp = server.request(
            method='PUT', user=user,
            path='/histomicsui/quarantine/%s/restore' % str(items[0]['_id']))
        assert utilities.respStatus(resp) == 403
        assert 'Write access denied' in resp.json['message']
        resp = server.request(
            method='PUT', user=admin,
            path='/histomicsui/quarantine/%s/restore' % str(items[0]['_id']))
        assert utilities.respStatus(resp) == 200
        resp = server.request(
            method='PUT', user=admin,
            path='/histomicsui/quarantine/%s/restore' % str(items[0]['_id']))
        assert utilities.respStatus(resp) == 400
        assert 'no quarantine record' in resp.json['message']
        resp = server.request(
            method='PUT', user=user,
            path='/histomicsui/quarantine/%s' % str(items[0]['_id']))
        assert utilities.respStatus(resp) == 200



import datetime
import json
import math
import time

import cachetools
import cherrypy
import girder.utility
import girder_large_image_annotation
import large_image.config
import orjson
from girder import logger
from girder.constants import AccessType
from girder.exceptions import RestException
from girder.models.file import File
from girder.models.folder import Folder
from girder.models.item import Item
from girder.models.setting import Setting
from girder.models.token import Token
from girder.models.user import User
from girder_large_image_annotation.models.annotation import Annotation

from .constants import PluginSettings

_recentIdentifiers = cachetools.TTLCache(maxsize=100, ttl=86400)


def _itemFromEvent(event, identifierEnding, itemAccessLevel=AccessType.READ):  # noqa
    """
    If an event has a reference and an associated identifier that ends with a
    specific string, return the associated item, user, and image file.

    :param event: the data.process event.
    :param identifierEnding: the required end of the identifier.
    :returns: a dictionary with item, user, and file if there was a match.
    """
    info = event.info
    identifier = None
    reference = info.get('reference', None)
    if reference is not None:
        try:
            reference = json.loads(reference)
            if (isinstance(reference, dict) and
                    isinstance(reference.get('identifier'), str)):
                identifier = reference['identifier']
        except (ValueError, TypeError):
            logger.debug('Failed to parse data.process reference: %r', reference)
    if identifier and 'uuid' in reference:
        if reference['uuid'] not in _recentIdentifiers:
            _recentIdentifiers[reference['uuid']] = {}
        _recentIdentifiers[reference['uuid']][identifier] = info
        reprocessFunc = _recentIdentifiers[reference['uuid']].pop('_reprocess', None)
        if reprocessFunc:
            reprocessFunc()
    if identifier is not None and identifier.endswith(identifierEnding):
        if 'itemId' not in reference and 'fileId' not in reference:
            logger.error('Reference does not contain at least one of itemId or fileId.')
            return
        userId = reference.get('userId')
        if not userId:
            if 'itemId' in reference:
                item = Item().load(reference['itemId'], force=True)
            else:
                file = File().load(reference['fileId'], force=True)
                item = Item().load(file['itemId'], force=True)
            if 'folderId' not in item:
                logger.error('Reference does not contain userId.')
                return
            folder = Folder().load(item['folderId'], force=True)
            userId = folder['creatorId']
        user = User().load(userId, force=True)
        imageId = reference.get('fileId')
        if not imageId:
            item = Item().load(reference['itemId'], force=True)
            if 'largeImage' in item and 'fileId' in item['largeImage']:
                imageId = item['largeImage']['fileId']
        image = File().load(imageId, level=AccessType.READ, user=user)
        item = Item().load(image['itemId'], level=itemAccessLevel, user=user)
        return {'item': item, 'user': user, 'file': image, 'uuid': reference.get('uuid')}


def resolveAnnotationGirderIds(event, results, data, possibleGirderIds):
    """
    If an annotation has references to girderIds, resolve them to actual ids.

    :param event: a data.process event.
    :param results: the results from _itemFromEvent,
    :param data: annotation data.
    :param possibleGirderIds: a list of annotation elements with girderIds
        needing resolution.
    :returns: True if all ids were processed.
    """
    # Exclude actual girderIds from resolution
    girderIds = []
    for element in possibleGirderIds:
        # This will throw an exception if the girderId isn't well-formed as an
        # actual id.
        try:
            if Item().load(element['girderId'], level=AccessType.READ, force=True) is None:
                girderIds.append(element)
        except Exception:
            girderIds.append(element)
    if not len(girderIds):
        return True
    idRecord = _recentIdentifiers.get(results.get('uuid'))
    if idRecord and not all(element['girderId'] in idRecord for element in girderIds):
        idRecord['_reprocess'] = lambda: process_annotations(event)
        return False
    for element in girderIds:
        element['girderId'] = str(idRecord[element['girderId']]['file']['itemId'])
        # Currently, all girderIds inside annotations are expected to be
        # large images.  In this case, load them and ask if they can be so,
        # in case they are small images
        from girder_large_image.models.image_item import ImageItem

        try:
            item = ImageItem().load(element['girderId'], force=True)
            ImageItem().createImageItem(
                item, list(ImageItem().childFiles(item=item, limit=1))[0], createJob=False)
        except Exception:
            pass
    return True


def process_annotations(event):  # noqa
    """Add annotations to an image on a ``data.process`` event"""
    results = _itemFromEvent(event, 'AnnotationFile')
    if not results:
        return
    item = results['item']
    user = results['user']

    file = File().load(
        event.info.get('file', {}).get('_id'),
        level=AccessType.READ, user=user,
    )
    startTime = time.time()

    if not file:
        logger.error('Could not load models from the database')
        return
    try:
        if file['size'] > int(large_image.config.getConfig(
                'max_annotation_input_file_length', 1024 ** 3)):
            msg = 'File is larger than will be read into memory.'
            raise Exception(msg)
        data = []
        with File().open(file) as fptr:
            while True:
                chunk = fptr.read(1024 ** 2)
                if not len(chunk):
                    break
                data.append(chunk)
        data = orjson.loads(b''.join(data).decode())
    except Exception:
        logger.error('Could not parse annotation file')
        raise
    if time.time() - startTime > 10:
        logger.info('Decoded json in %5.3fs', time.time() - startTime)

    if not isinstance(data, list) or (
            hasattr(girder_large_image_annotation.utils, 'isGeoJSON') and
            girder_large_image_annotation.utilsisGeoJSON(data)):
        data = [data]
    # Check some of the early elements to see if there are any girderIds
    # that need resolution.
    if 'uuid' in results:
        girderIds = [
            element for annotation in data
            for element in annotation.get('elements', [])[:100]
            if 'girderId' in element]
        if len(girderIds):
            if not resolveAnnotationGirderIds(event, results, data, girderIds):
                return
    for annotation in data:
        try:
            Annotation().createAnnotation(item, user, annotation)
        except Exception:
            logger.error('Could not create annotation object from data')
            raise
    if Setting().get(PluginSettings.HUI_DELETE_ANNOTATIONS_AFTER_INGEST):
        item = Item().load(file['itemId'], force=True)
        if item and len(list(Item().childFiles(item, limit=2))) == 1:
            Item().remove(item)


def quarantine_item(item, user, makePlaceholder=True):
    """
    Quarantine an item, marking which user did it.  Note that this raises
    RestExceptions for failures.

    :param user: the user doing the quarantining.
    :param item: an item to quarantine.
    :returns: the modified item.
    """
    folder = Setting().get(PluginSettings.HUI_QUARANTINE_FOLDER)
    if not folder:
        msg = 'The quarantine folder is not configured.'
        raise RestException(msg)
    folder = Folder().load(folder, force=True, exc=True)
    if not folder:
        msg = 'The quarantine folder does not exist.'
        raise RestException(msg)
    if str(folder['_id']) == str(item['folderId']):
        msg = 'The item is already in the quarantine folder.'
        raise RestException(msg)
    originalFolder = Folder().load(item['folderId'], force=True)
    quarantineInfo = {
        'originalFolderId': item['folderId'],
        'originalBaseParentType': item['baseParentType'],
        'originalBaseParentId': item['baseParentId'],
        'originalUpdated': item['updated'],
        'quarantineUserId': user['_id'],
        'quarantineTime': datetime.datetime.now(datetime.timezone.utc),
    }
    item = Item().move(item, folder)
    if makePlaceholder:
        placeholder = Item().createItem(
            item['name'] + ' [Removed - Quarantined]',
            {'_id': item['creatorId']}, originalFolder,
            description=item['description'])
        quarantineInfo['placeholderItemId'] = placeholder['_id']
    item.setdefault('meta', {})['quarantine'] = quarantineInfo
    item = Item().updateItem(item)
    if makePlaceholder:
        placeholderInfo = {
            'quarantined': True,
            'quarantineTime': quarantineInfo['quarantineTime'],
        }
        placeholder.setdefault('meta', {})['quarantine'] = placeholderInfo
        placeholder = Item().updateItem(placeholder)
    return item


def restore_quarantine_item(item, user):
    """
    Unquarantine an item, returning it to its original location.  Note that
    this raises RestExceptions for failures.

    :param item: an item to unquarantine.
    :returns: the modified item.
    """
    if not item.get('meta', {}).get('quarantine'):
        msg = 'The item has no quarantine record.'
        raise RestException(msg)
    folder = Folder().load(item['meta']['quarantine']['originalFolderId'], force=True)
    if not folder:
        msg = 'The original folder is not accessible.'
        raise RestException(msg)
    if 'placeholderItemId' in item['meta']['quarantine']:
        placeholder = Item().load(item['meta']['quarantine']['placeholderItemId'], force=True)
    else:
        placeholder = None
    item = Item().move(item, folder)
    item['updated'] = item['meta']['quarantine']['originalUpdated']
    del item['meta']['quarantine']
    item = Item().updateItem(item)
    if placeholder is not None:
        Item().remove(placeholder)
    return item


def process_metadata(event):
    """Add metadata to an item on a ``data.process`` event"""
    results = _itemFromEvent(event, 'ItemMetadata', AccessType.WRITE)
    if not results:
        return
    file = File().load(
        event.info.get('file', {}).get('_id'),
        level=AccessType.READ, user=results['user'],
    )

    if not file:
        logger.error('Could not load models from the database')
        return
    try:
        data = orjson.loads(File().open(file).read().decode())
    except Exception:
        logger.error('Could not parse metadata file')
        raise

    item = results['item']
    Item().setMetadata(item, data, allowNull=False)


def nan2None(obj):
    """
    Convert NaN and +/-Infitity to None.
    """
    if isinstance(obj, dict):
        return {k: nan2None(v) for k, v in obj.items()}
    elif isinstance(obj, list):
        return [nan2None(v) for v in obj]
    elif isinstance(obj, float) and (
            math.isnan(obj) or obj == float('inf') or obj == -float('inf')):
        print('Nan')
        return None
    return obj


def json_nans_as_nulls():
    """
    Convert NaN and +/-Infinity to nulls so that they will serialize
    """

    def encode(self, obj, *args, **kwargs):
        obj = nan2None(obj)
        return json.JSONEncoder.encode(self, obj, *args, **kwargs)

    def iterencode(self, obj, *args, **kwargs):
        obj = nan2None(obj)
        return json.JSONEncoder.iterencode(self, obj, *args, **kwargs)

    girder.utility.JsonEncoder.encode = encode
    girder.utility.JsonEncoder.iterencode = iterencode


def shortLoginSessions():
    import girder.api.rest

    _recentTokens = {}

    origGetCurrentUser = girder.api.rest.getCurrentUser

    def getCurrentUser(*args, **kwargs):
        result = origGetCurrentUser(*args, **kwargs)
        try:
            if 'api/v1/notification/stream' in cherrypy.request.path_info:
                return result
        except Exception:
            pass
        user = result[0] if isinstance(result, tuple) else result
        if user:
            token = girder.api.rest.getCurrentToken()
        if user and token:
            if token['_id'] not in _recentTokens or time.time() - _recentTokens[token['_id']] > 60:
                if Setting().get(PluginSettings.HUI_LOGIN_SESSION_EXPIRY_MINUTES):
                    days = float(Setting().get(
                        PluginSettings.HUI_LOGIN_SESSION_EXPIRY_MINUTES)) / 24 / 60
                    token['expires'] = datetime.datetime.utcnow() + datetime.timedelta(
                        days=float(days))
                    token = Token().save(token)
                    logger.debug(
                        'Extend user login duration '
                        f'(user {user["_id"]}, token {token["_id"][:16]}...)')
            if len(_recentTokens) > 100:
                _recentTokens.clear()
            _recentTokens[token['_id']] = time.time()
        return result

    girder.api.rest.getCurrentUser = getCurrentUser

    origResourceSendAuthTokenCookie = girder.api.rest.Resource.sendAuthTokenCookie

    def sendAuthTokenCookie(self, user=None, scope=None, token=None, days=None):
        if days is None and Setting().get(PluginSettings.HUI_LOGIN_SESSION_EXPIRY_MINUTES):
            days = float(Setting().get(PluginSettings.HUI_LOGIN_SESSION_EXPIRY_MINUTES)) / 24 / 60
        return origResourceSendAuthTokenCookie(self, user, scope, token, days)

    girder.api.rest.Resource.sendAuthTokenCookie = origResourceSendAuthTokenCookie



import mimetypes
import re

from girder.constants import AccessType
from girder.exceptions import ValidationException
from girder.models.collection import Collection
from girder.models.file import File
from girder.models.folder import Folder
from girder.models.item import Item
from girder.models.setting import Setting
from girder.utility.model_importer import ModelImporter

from ..constants import TCGACollectionSettingKey


def pruneNoneValues(d):
    """Recursively prune dictionary items with value `None`."""
    toDelete = [k for k, v in d.items() if v is None]
    for k in toDelete:
        del d[k]
    for _, v in d.items():
        if isinstance(v, dict):
            pruneNoneValues(v)


def updateDict(d, u):
    """Recursively update a dictionary with items from another."""
    for k, v in u.items():
        if isinstance(v, dict):
            r = updateDict(d.get(k, {}), v)
            d[k] = r
        else:
            d[k] = u[k]
    return d


class TCGAModel:
    """
    This class is used as a mixin for all TCGA model classes.  TCGA models
    are distinguished by the existence of a `tcga` property at the top level
    of their document.  Each model can use this key differently, but all
    models must set `tcga.type` to be the model subclass.  This is handled
    automatically by the mixin from the TCGAType static call property.
    This mixin overrides several Girder core methods to properly set tcga
    metadata as well as limit queries to only the given model type.

    The special `tcga` object contained in these documents can contain an
    additional `meta` field which is interpreted as model specific metadata.
    This is analogous to the `meta` field in Girder core items and folders.
    """

    #: Additional indices to add to the collection
    TCGAIndices = []

    #: Matches valid case names
    case_re = re.compile(r'tcga-[a-z0-9]{2}-[a-z0-9]{4}', flags=re.I)

    #: Matches valid uuid's
    uuid_re = re.compile(
        r'[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}',
        flags=re.I,
    )

    #: Matches valid tcga barcodes
    barcode_re = re.compile(
        case_re.pattern + r'[0-9a-z-]*',
        flags=re.I,
    )

    #: Parses tcga slide image file names
    image_re = re.compile(
        r'^(?P<barcode>(?P<case>' +
        case_re.pattern + r')[0-9a-z-]*)\.' +
        r'(?P<uuid>' + uuid_re.pattern + r')\.svs$',
        flags=re.I,
    )

    #: Parses tcga pathology report file names
    pathology_re = re.compile(
        r'^(?P<case>' + case_re.pattern +
        r')\.(?P<uuid>' + uuid_re.pattern +
        r')\.pdf$',
        flags=re.I,
    )

    def initialize(self, **kwargs):
        """Expose the tcga key as public metadata."""
        self.exposeFields(AccessType.READ, fields='tcga')
        self.ensureIndices(['tcga.type'] + list(self.TCGAIndices))
        super().initialize(**kwargs)
        self.name = self.TCGAType

    def save(self, doc, baseModel=False, **kwargs):
        """Set the TCGA model type on save."""
        if not baseModel:
            self.setTCGA(doc, type=self.TCGAType)
        return super().save(doc, **kwargs)

    def find(self, query=None, **kwargs):
        """Append TCGA model type to any query on this model."""
        query = query or {}
        query['tcga.type'] = self.TCGAType
        return super().find(query, **kwargs)

    def findOne(self, query=None, **kwargs):
        """Append TCGA model type to any query on this model."""
        query = query or {}
        query['tcga.type'] = self.TCGAType
        return super().findOne(query, **kwargs)

    def setTCGA(self, doc, **tcga):
        """Update the TCGA object and prune values of None."""
        self.getTCGA(doc).update(tcga)
        pruneNoneValues(self.getTCGA(doc))
        return self

    def getTCGA(self, doc):
        """Return the TCGA object."""
        return doc.setdefault('tcga', {})

    def updateTCGAMeta(self, doc, meta):
        """Update TCGA metadata."""
        meta = updateDict(self.getTCGAMeta(doc), meta)
        pruneNoneValues(meta)
        return self

    def getTCGAMeta(self, doc):
        """Return TCGA metadata from a document."""
        tcga = self.getTCGA(doc)
        return tcga.setdefault('meta', {})

    def removeTCGA(self, doc):
        """Remove the tcga key and save the document.

        This method will effectively reset the document so that
        it no longer behaves as a TCGA document.  This is the
        opposite of the import methods that promote Girder models
        to TCGA types.
        """
        if 'tcga' in doc:
            del doc['tcga']
            self.save(doc, baseModel=True, validate=False)
        return doc

    def getTCGAType(self, doc):
        """Get the type of model expressed by the document."""
        return self.getTCGA(doc).get('type')

    def getTCGACollection(self):
        """Get the unique TCGA collection from the settings collection."""
        tcga = Setting().get(
            TCGACollectionSettingKey,
        )
        if tcga is None:
            msg = 'TCGA collection id not initialized in settings'
            raise Exception(
                msg,
            )
        return Collection().load(
            tcga, force=True,
        )

    def __upper(self, obj, key):
        if key in obj:
            obj[key] = obj[key].upper()

    def __lower(self, obj, key):
        if key in obj:
            obj[key] = obj[key].lower()

    def _parse(self, name, regex):
        m = regex.match(name)
        if m is None:
            msg = 'Invalid name'
            raise ValidationException(msg)
        d = m.groupdict()
        self.__lower(d, 'uuid')
        self.__upper(d, 'barcode')
        self.__upper(d, 'case')
        return d

    def parseImage(self, name):
        """Parse a slide image file name."""
        return self._parse(name, self.image_re)

    def parsePathology(self, name):
        """Parse a pathology report file name."""
        return self._parse(name, self.pathology_re)

    @classmethod
    def setMimeType(cls, doc):
        """Set the mime type of a file document."""
        oldType = doc.get('mimeType')
        newType = mimetypes.guess_type(
            doc.get('name', ''),
        )[0]
        doc['mimeType'] = newType
        return newType != oldType

    def importDocument(self, doc, **kwargs):
        """Promote a Girder core document to a TCGA model."""
        self.setTCGA(doc)
        self.save(doc)
        if doc.get('_modelType') == 'item':
            fileModel = File()
            files = Item().childFiles(doc)
            for file in files:
                if self.setMimeType(file):
                    fileModel.save(file)
        return doc

    def loadDocument(self, id, **kwargs):
        """Load a user, folder, item, or collection document."""
        for modelType in ('collection', 'user', 'folder', 'item'):
            model = ModelImporter.model(modelType)
            try:
                doc = model.load(id, **kwargs)
                if doc:
                    doc['_modelType'] = modelType
                    return doc
            except ValidationException:
                pass
        msg = 'Invalid document id provided'
        raise ValidationException(
            msg,
        )

    def iterateItems(self, doc, **kwargs):
        """Iterate over all items under the given document."""
        folder = Folder()
        item = Item()
        for child in item.find({'folderId': doc['_id']}):
            yield child

        for child in folder.find({'parentId': doc['_id']}):
            yield from self.iterateItems(child, **kwargs)

    def childFolders(self, parent, parentType, user=None, limit=0, offset=0,
                     sort=None, filters=None, cursor=False, **kwargs):
        """Add a ``cursor`` option to the standard childFolders method.

        The cursor option will ensure the response is a mongo cursor rather
        than a generic generator.  The resulting cursor will *not* be
        filtered by access control.  This option exists to allow efficient
        paging for resources that don't require fine grained permissions.
        """
        if not cursor:
            return super().childFolders(
                parent, parentType, limit=limit, offset=offset, sort=sort, **kwargs)

        parentType = parentType.lower()
        q = {
            'parentId': parent['_id'],
            'parentCollection': parentType,
        }
        q.update(filters or {})
        return self.find(q, sort=sort, limit=limit, offset=offset)



import os

from setuptools import find_packages, setup

with open('README.rst') as readme_file:
    readme = readme_file.read()


def prerelease_local_scheme(version):
    """
    Return local scheme version unless building on master in CircleCI.

    This function returns the local scheme version number
    (e.g. 0.0.0.dev<N>+g<HASH>) unless building on CircleCI for a
    pre-release in which case it ignores the hash and produces a
    PEP440 compliant pre-release version number (e.g. 0.0.0.dev<N>).
    """
    from setuptools_scm.version import get_local_node_and_date

    if os.getenv('CIRCLE_BRANCH') in ('master', ):
        return ''
    else:
        return get_local_node_and_date(version)


setup(
    name='histomicsui',
    use_scm_version={'local_scheme': prerelease_local_scheme, 'fallback_version': '0.0.0'},
    setup_requires=['setuptools-scm'],
    description='Organize, visualize, and analyze histology images.',
    author='Kitware, Inc.',
    author_email='kitware@kitware.com',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
    ],
    install_requires=[
        'girder-large-image-annotation>=1.31.0',
        'girder-slicer-cli-web[girder]>=1.4.0',
        'cachetools',
        'orjson',
    ],
    extras_require={
        'analysis': [],  # kept for backwards compatibility
    },
    license='Apache Software License 2.0',
    long_description=readme,
    long_description_content_type='text/x-rst',
    include_package_data=True,
    keywords='girder-plugin, histomicsui',
    packages=find_packages(exclude=['test', 'test.*']),
    url='https://github.com/DigitalSlideArchive/histomicsui',
    zip_safe=False,
    python_requires='>=3.8',
    entry_points={
        'girder.plugin': [
            'histomicsui = histomicsui:GirderPlugin',
        ],
    },
)



import re

from girder.constants import AccessType
from girder.models.file import File
from girder.models.item import Item
from girder.models.model_base import ValidationException

from .image import Image as TCGAImage
from .meta import TCGAModel


class Aperio(Item):

    #: Parses tcga Aperio annotation file names as tcga types
    aperio_re = re.compile(
        r'^(?P<barcode>(?P<case>' +
        TCGAModel.case_re.pattern + r')[0-9a-z-]*)\.xml',
        flags=re.I,
    )

    def initialize(self, **kwargs):
        """Expose aperio key as public metadata."""
        self.exposeFields(AccessType.READ, fields='aperio')
        self.ensureIndices(['aperio.image', 'aperio.tag'])
        return super().initialize(**kwargs)

    def validate(self, doc, **kwargs):
        """Ensure the item has valid metadata."""
        super().validate(doc, **kwargs)
        meta = doc.setdefault('aperio', {})
        imageId = meta.get('image')
        try:
            Item().load(imageId, force=True)
        except Exception:
            msg = 'The item is not associated with a valid image.'
            raise ValidationException(
                msg,
            )

        files = Item().childFiles(doc)
        if files.count() != 1:
            msg = 'The annotation item must have exactly one file.'
            raise ValidationException(
                msg,
            )

        meta.setdefault('tag', None)
        return doc

    def setTag(self, doc, tag):
        meta = doc.setdefault('aperio', {})
        meta['tag'] = tag
        self.save(doc)
        return doc

    def importDocument(self, doc, image, **kwargs):
        """Promote the item document to an aperio type."""
        tag = kwargs.get('tag')
        meta = doc.setdefault('aperio', {})
        meta['image'] = image['_id']
        self.setTag(doc, tag)
        fileModel = File()
        files = Item().childFiles(doc)
        for file in files:
            if TCGAModel.setMimeType(file):
                fileModel.save(file)
        return doc

    def importTCGADocument(self, doc, **kwargs):
        """Import the annotation item into a tcga structure."""
        recurse = kwargs.get('recurse', False)

        if doc.get('_modelType', 'item') == 'item':
            name = doc['name']
            d = self.aperio_re.match(name).groupdict()
            image = TCGAImage().findOne({
                'tcga.barcode': d['barcode'].upper(),
            })
            self.importDocument(doc, image, **kwargs)

            return doc
        elif recurse:
            for item in self.iterateItems(doc):
                try:
                    self.importTCGADocument(item, **kwargs)
                except ValidationException:
                    pass
        else:
            msg = 'Invalid model type'
            raise ValidationException(msg)

    def findAperio(self, image, tag=None, **kwargs):
        """Find Aperio annotations associated with the given image."""
        query = {
            'aperio.image': image['_id'],
        }
        if tag is not None:
            query['aperio.tag'] = tag
        return self.find(query, **kwargs)

    def removeAperio(self, doc):
        """Remove Aperio metadata from an item."""
        if 'aperio' in doc:
            del doc['aperio']
            Item().save(doc)
        return doc



import os
import shutil

import pytest
from girder.api import access
from girder.api.describe import Description, describeRoute
from girder.api.rest import Resource, setRawResponse, setResponseHeader
from girder.constants import STATIC_ROOT_DIR
from girder.models.folder import Folder
from girder.models.item import Item
from girder_large_image.models.image_item import ImageItem
from girder_large_image_annotation.models.annotation import Annotation
from pytest_girder.web_client import runWebClientTest

from . import girder_utilities as utilities


def copyHUITest():
    src = os.path.join(os.path.dirname(__file__), 'web_client_specs', 'huiTest.js')
    dest = os.path.join(STATIC_ROOT_DIR, 'built/plugins/histomicsui', 'huiTest.js')
    if not os.path.exists(dest) or os.path.getmtime(src) != os.path.getmtime(dest):
        shutil.copy2(src, dest)


def makeResources(server, fsAssetstore, admin, user):
    # Create an item in the admin Public folder
    adminPublicFolder = Folder().childFolders(  # noqa: B305
        admin, 'user', filters={'name': 'Public'},
    ).next()
    Item().createItem('Empty', admin, adminPublicFolder)
    # Upload a sample file
    file = utilities.uploadExternalFile(
        'sample_svs_image.TCGA-DU-6399-01A-01-TS1.e8eb65de-d63e-42db-af6f-14fefbbdf7bd.svs',
        user, fsAssetstore, name='image')
    item = Item().load(file['itemId'], force=True)
    # We have to ask to make this a large image item, because we renamed it
    # 'image' without an extension
    ImageItem().createImageItem(item, file, user=user, createJob=False)
    ImageItem().copyItem(item, user, name='copy')

    annotation = Annotation().createAnnotation(item, admin, {'name': 'admin annotation'})
    annotation = Annotation().setAccessList(annotation, {}, force=True, save=False)
    annotation = Annotation().setPublic(annotation, True, save=True)

    Folder().createFolder(user, 'subfolder', parentType='user', public=True, creator=user)
    utilities.uploadTestFile('.histomicsui_config.yaml', user, fsAssetstore, 'subfolder')
    utilities.uploadExternalFile(
        'sample_svs_image.TCGA-DU-6399-01A-01-TS1.e8eb65de-d63e-42db-af6f-14fefbbdf7bd.svs',
        user, fsAssetstore, folderName='subfolder', name='image.svs')


class MockSlicerCLIWebResource(Resource):
    """
    This creates a mocked version of the ``/HistomicsTK/HistomicsTK/docker_image``
    endpoint so we can test generation of the analysis panel on the client without
    relying on girder_worker + docker.
    """

    def __init__(self):
        super().__init__()
        self.route('GET', ('docker_image',), self.dockerImage)
        self.route('GET', ('test_analysis_detection', 'xml'), self.testAnalysisXmlDetection)
        self.route('GET', ('test_analysis_features', 'xml'), self.testAnalysisXmlFeatures)
        self.route('POST', ('test_analysis_detection', 'run'), self.testAnalysisRun)
        self.route('POST', ('test_analysis_features', 'run'), self.testAnalysisRun)

    @access.public
    @describeRoute(
        Description('Mock the docker_image endpoint.'),
    )
    def dockerImage(self, params):
        """
        Return a single CLI referencing mocked out /xmlspec and /run endpoints.
        """
        return {
            'huirchive/histomicstk': {
                'latest': {
                    'ComputeNucleiFeatures': {
                        'run': 'mock_resource/test_analysis_features/run',
                        'type': 'python',
                        'xmlspec': 'mock_resource/test_analysis_features/xml',
                    },
                    'NucleiDetection': {
                        'run': 'mock_resource/test_analysis_detection/run',
                        'type': 'python',
                        'xmlspec': 'mock_resource/test_analysis_detection/xml',
                    },
                },
            },
        }

    @access.public
    @describeRoute(
        Description('Mock an analysis description route.'),
    )
    def testAnalysisXmlDetection(self, params):
        """Return the nuclei detection XML spec as a test case."""
        xml_file = os.path.abspath(os.path.join(
            os.path.dirname(__file__), 'test_files', 'test_analysis_detection.xml'))
        with open(xml_file) as f:
            xml = f.read()
        setResponseHeader('Content-Type', 'application/xml')
        setRawResponse()
        return xml

    @access.public
    @describeRoute(
        Description('Mock an analysis description route.'),
    )
    def testAnalysisXmlFeatures(self, params):
        """Return the nuclei feature classification XML spec as a test case."""
        xml_file = os.path.abspath(os.path.join(
            os.path.dirname(__file__), 'test_files', 'test_analysis_features.xml'))
        with open(xml_file) as f:
            xml = f.read()
        setResponseHeader('Content-Type', 'application/xml')
        setRawResponse()
        return xml

    @access.public
    @describeRoute(
        Description('Mock an analysis run route.'),
    )
    def testAnalysisRun(self, params):
        """
        Mock out the CLI execution endpoint.

        For now, this is a no-op, but we should add some logic to generate an annotation
        output and job status events to simulate a real execution of the CLI.
        """
        return {'_id': 'jobid'}


@pytest.mark.plugin('histomicsui')
@pytest.mark.parametrize('spec', [
    'analysisSpec.js',
    'annotationSpec.js',
    'girderUISpec.js',
    'huiSpec.js',
    'itemSpec.js',
    'metadataPanelSpec.js',
    'metadataPlotSpec.js',
    'overviewPanelSpec.js',
    'panelLayoutSpec.js',
    'pixelmapCategorySpec.js',
])
def testWebClient(boundServer, fsAssetstore, db, admin, user, spec):
    copyHUITest()
    boundServer.root.api.v1.mock_resource = MockSlicerCLIWebResource()
    makeResources(boundServer, fsAssetstore, admin, user)
    spec = os.path.join(os.path.dirname(__file__), 'web_client_specs', spec)
    runWebClientTest(boundServer, spec, 15000)
